# coding: utf-8
# Copyright 2016 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
**Module to compute beam slicing for a sparse beam**
**Only valid for cases with constant revolution and RF frequencies**

:Authors: **Juan F. Esteban Mueller**
'''

from __future__ import division, print_function

from builtins import range

import numpy as np

from ..beam.profile import CutOptions, Profile
from ..utils import bmath as bm


class SparseSlices:
    '''
    *This class instantiates a Profile object for each filled bucket according
    to the provided filling pattern. Each Profile object will be of the size of
    an RF bucket and will have the same number of slices.*
    '''

    def __init__(self, RFStation, Beam, n_slices_bucket, filling_pattern, tracker='C',
                 direct_slicing=False):

        #: *Import (reference) Beam*
        self.Beam = Beam

        #: *Import (reference) RFStation*
        self.RFParams = RFStation

        #: *Number of slices per bucket*
        self.n_slices_bucket = n_slices_bucket

        #: *Filling pattern as a boolean array where True (1) means filled
        # bucket*
        self.filling_pattern = filling_pattern

        # Bunch index for each filled bucket (-1 if empty). Only for C++ track
        self.bunch_indexes = np.cumsum(filling_pattern) * filling_pattern - 1

        #: *Number of buckets to be sliced*
        self.n_filled_buckets = int(np.sum(filling_pattern))

        # Pre-processing the slicing edges
        self.set_cuts()

        # Initialize individual slicing objects
        self.profiles_list = []
        # Group n_macroparticles from all objects in a single array
        # (for C++ track).
        self.n_macroparticles_array = np.zeros((self.n_filled_buckets,
                                                n_slices_bucket))
        # Group bin_centers from all objects in a single array (for impedance)
        self.bin_centers_array = np.zeros((self.n_filled_buckets, n_slices_bucket))
        self.edges_array = np.zeros((self.n_filled_buckets, n_slices_bucket + 1))
        for i in range(self.n_filled_buckets):
            # Only valid for cut_edges='edges'

            self.profiles_list.append(Profile(Beam, CutOptions(cut_left=self.cut_left_array[i],
                                                               cut_right=self.cut_right_array[i],
                                                               n_slices=n_slices_bucket)))

            self.profiles_list[i].n_macroparticles = self.n_macroparticles_array[i, :]
            self.bin_centers_array[i, :] = self.profiles_list[i].bin_centers
            self.edges_array[i, :] = self.profiles_list[i].edges
            self.profiles_list[i].bin_centers = self.bin_centers_array[i, :]

        # Select the tracker
        if tracker == 'C':
            self.track = self._histogram_c
        elif tracker == 'onebyone':
            self.track = self._histrogram_one_by_one
        else:
            # WrongCalcError
            raise RuntimeError(
                'Tracking method not recognized!')

        # Track at initialisation
        if direct_slicing:
            self.track()

    def set_cuts(self):
        '''
        *Method to set the self.cut_left_array and self.cut_right_array
        properties, with the limits being an RF period.
        This is done as a pre-processing.*
        '''
        # RF period
        t_rf = self.RFParams.t_rf[0, self.RFParams.counter[0]]

        self.cut_left_array = np.zeros(self.n_filled_buckets)
        self.cut_right_array = np.zeros(self.n_filled_buckets)
        for i in range(self.n_filled_buckets):
            bucket_index = np.where(self.filling_pattern)[0][i]
            self.cut_left_array[i] = bucket_index * t_rf
            self.cut_right_array[i] = (bucket_index + 1) * t_rf

    def _histogram_c(self):
        '''
        *Histrogram generated by calling an optimized C++ function that
        calculates all the profile at once.*
        '''
        bm.sparse_histogram(self.Beam.dt, self.n_macroparticles_array,
                            self.cut_left_array, self.cut_right_array,
                            self.bunch_indexes, self.n_slices_bucket)

    def _histrogram_one_by_one(self):
        '''
        *Histrogram generated by calling the tack() method of each Profile
        object*
        '''

        for i in range(self.n_filled_buckets):
            self.profiles_list[i].track()




'''
**Optimised next_regular routine to generate Hamming numbers: 
the output will contain at least one factor 2.**

:Authors: **Alexandre Lasheen**, **Juan F. Esteban Mueller**, **Danilo Quartullo** 
'''


def next_regular(target):
    """
    Find the next regular number greater than or equal to target.
    Regular numbers are composites of the prime factors 2, 3, and 5.
    Also known as 5-smooth numbers or Hamming numbers, these are the optimal
    size for inputs to FFTPACK.

    Target must be a positive integer.
    """
    if target <= 6:
        return target

    # Quickly check if it's already a power of 2
    if not (target & (target - 1)):
        return target

    target = -(-target // 2)

    match = float('inf')  # Anything found will be smaller
    p5 = 1
    while p5 < target:
        p35 = p5
        while p35 < target:
            # Ceiling integer division, avoiding conversion to float
            # (quotient = ceil(target / p35))
            quotient = -(-target // p35)

            # Quickly find next power of 2 >= quotient
            try:
                p2 = 2**((quotient - 1).bit_length())
            except AttributeError:
                # Fallback for Python <2.7
                p2 = 2**(len(bin(quotient - 1)) - 2)

            N = p2 * p35
            if N == target:
                return N * 2
            elif N < match:
                match = N
            p35 *= 3
            if p35 == target:
                return p35 * 2
        if p35 < match:
            match = p35
        p5 *= 5
        if p5 == target:
            return p5 * 2
    if p5 < match:
        match = p5

    return match * 2

# THIS IS THE ORIGINAL PYTHON VERSION #

# def next_regular(target):
#     """
#     Find the next regular number greater than or equal to target.
#     Regular numbers are composites of the prime factors 2, 3, and 5.
#     Also known as 5-smooth numbers or Hamming numbers, these are the optimal
#     size for inputs to FFTPACK.
#
#     Target must be a positive integer.
#     """
#     if target <= 6:
#         return target
#
#     # Quickly check if it's already a power of 2
#     if not (target & (target-1)):
#         return target
#
#     match = float('inf') # Anything found will be smaller
#     p5 = 1
#     while p5 < target:
#         p35 = p5
#         while p35 < target:
#             # Ceiling integer division, avoiding conversion to float
#             # (quotient = ceil(target / p35))
#             quotient = -(-target // p35)
#
#             # Quickly find next power of 2 >= quotient
#             try:
#                 p2 = 2**((quotient - 1).bit_length())
#             except AttributeError:
#                 # Fallback for Python <2.7
#                 p2 = 2**(len(bin(quotient - 1)) - 2)
#
#             N = p2 * p35
#             if N == target:
#                 return N
#             elif N < match:
#                 match = N
#             p35 *= 3
#             if p35 == target:
#                 return p35
#         if p35 < match:
#             match = p35
#         p5 *= 5
#         if p5 == target:
#             return p5
#     if p5 < match:
#         match = p5
#
#     return match




# Copyright 2016 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
**Module to plot different LLRF features**

:Authors: **Helga Timko**, **Danilo Quartullo**
'''

from __future__ import division

from builtins import range

import matplotlib.pyplot as plt
import numpy as np
from pylab import cm



def plot_noise_spectrum(frequency, spectrum, sampling = 1, dirname = 'fig', show_plot = False,
                        figno = 0):
    
    """
    Plot of the phase noise spectrum.
    For large amount of data, use "sampling" to plot a fraction of the data.
    """

    # Plot
    fig = plt.figure(1)
    fig.set_size_inches(8,6)
    ax = plt.axes()
    ax.set_xlim([0, 300])    
    ax.plot(frequency[::sampling], spectrum[::sampling])
    ax.set_xlabel("Frequency [Hz]")
    params = {'text.usetex': False, 'mathtext.default': 'sf'}
    plt.rcParams.update(params)
    ax.set_ylabel(r"Noise spectrum [$\frac{rad^2}{Hz}$]")
    ax.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))

    # Output plot
    if show_plot:
        plt.show()
    else:
        fign = dirname +'/noise_spectrum_' "%d" %figno +'.png'
        plt.savefig(fign)
    plt.clf()

        
    
def plot_phase_noise(time, dphi, sampling = 1, dirname = 'fig', show_plot = False,
                     figno = 0):
    
    """
    Plot of phase noise as a function of time.
    For large amount of data, use "sampling" to plot a fraction of the data.
    """

    # Plot
    fig = plt.figure(1)
    fig.set_size_inches(8,6)
    ax = plt.axes()
    ax.plot(time[::sampling], dphi[::sampling])
    ax.set_xlabel("Time [s]")
    ax.set_ylabel(r"Phase noise [rad]")

    # Output plot
    if show_plot:
        plt.show()
    else:
        fign = dirname +'/phase_noise_' "%d" %figno +'.png'
        plt.savefig(fign)
    plt.clf()     


def plot_PL_bunch_phase(RFStation, h5data, output_freq = 1, 
                        dirname = 'fig', show_plot = False):
    
    """
    Plot of bunch phase measured by the Phase Loop as a function of time.
    For large amount of data, monitor with larger 'output_freq'.
    """

    # Time step of plotting
    time_step = RFStation.counter[0]

    # Load/create data
    if output_freq < 1:
        output_freq = 1
    ndata = int(time_step / output_freq)
    t = output_freq * np.arange(ndata)
    dphi = np.array(h5data["/Beam/PL_bunch_phase"][0:ndata], dtype=np.double)
    dphi[time_step:] = np.nan

    # Plot
    fig = plt.figure(1)
    fig.set_size_inches(8,6)
    ax = plt.axes()
    ax.plot(t, dphi,'.')
    ax.set_xlabel(r"No. turns [T$_0$]")    
    ax.set_ylabel (r"PL $\phi_{\mathsf{bunch}}$ [rad]")
    ax.ticklabel_format(style='sci', axis='y', scilimits=(0,0))
    if time_step > 100000:
        ax.ticklabel_format(style='sci', axis='x', scilimits=(0, 0))

    # Output plot
    if show_plot:
        plt.show()
    else:
        fign = dirname +'/PL_bunch_phase.png'
        plt.savefig(fign)
    plt.clf()     


def plot_PL_RF_phase(RFStation, h5data, output_freq = 1, 
                    dirname = 'fig', show_plot = False):
    
    """
    Plot of RF phase; monitored with Phase Loop.
    For large amount of data, monitor with larger 'output_freq'.
    """

    # Time step of plotting
    time_step = RFStation.counter[0]

    # Load/create data
    if output_freq < 1:
        output_freq = 1
    ndata = int(time_step / output_freq)
    t = output_freq * np.arange(ndata)
    dphi = np.array(h5data["/Beam/PL_phiRF"][0:ndata], dtype=np.double)
    dphi[time_step:] = np.nan

    # Plot
    plt.figure(1, figsize=(8,6))
    ax = plt.axes()
    ax.plot(t, dphi, '.')
    ax.set_xlabel(r"No. turns [T$_0$]")
    ax.set_ylabel(r"RF phase $\phi_{\mathsf{RF}}$ [rad]")
    ax.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
    if time_step > 100000:
        ax.ticklabel_format(style='sci', axis='x', scilimits=(0, 0))

    # Output plot
    if show_plot:
        plt.show()
    else:
        fign = dirname +'/RF_phase.png'
        plt.savefig(fign)
    plt.clf()     


def plot_PL_phase_corr(RFStation, h5data, output_freq = 1, 
                       dirname = 'fig', show_plot = False):
    
    """
    Plot of phase correction applied by the Phase Loop as a function of time.
    For large amount of data, monitor with larger 'output_freq'.
    """

    # Time step of plotting
    time_step = RFStation.counter[0]

    # Load/create data
    if output_freq < 1:
        output_freq = 1
    ndata = int(time_step / output_freq)
    t = output_freq * np.arange(ndata)
    dphi = np.array(h5data["/Beam/PL_phase_corr"][0:ndata], dtype=np.double)
    dphi[time_step:] = np.nan

    # Plot
    fig = plt.figure(1)
    fig.set_size_inches(8,6)
    ax = plt.axes()
    ax.plot(t, dphi,'.')
    ax.set_xlabel(r"No. turns [T$_0$]")    
    ax.set_ylabel (r"PL $\phi$ correction [rad]")
    ax.ticklabel_format(style='sci', axis='y', scilimits=(0,0))
    if time_step > 100000:
        ax.ticklabel_format(style='sci', axis='x', scilimits=(0, 0))

    # Output plot
    if show_plot:
        plt.show()
    else:
        fign = dirname +'/PL_phase_corr.png'
        plt.savefig(fign)
    plt.clf()     


def plot_PL_RF_freq(RFStation, h5data, output_freq = 1, 
                    dirname = 'fig', show_plot = False):
    
    """
    Plot of RF revolution frequency; monitored with Phase Loop.
    For large amount of data, monitor with larger 'output_freq'.
    """

    # Time step of plotting
    time_step = RFStation.counter[0]

    # Load/create data
    if output_freq < 1:
        output_freq = 1
    ndata = int(time_step / output_freq)
    t = output_freq * np.arange(ndata)
    dphi = np.array(h5data["/Beam/PL_omegaRF"][0:ndata], dtype=np.double)
    dphi[time_step:] = np.nan

    # Plot
    plt.figure(1, figsize=(8,6))
    ax = plt.axes()
    ax.plot(t, dphi, '.')
    ax.set_xlabel(r"No. turns [T$_0$]")
    ax.set_ylabel(r"RF revolution frequency $\omega_{\mathsf{RF}}$ [1/s]")
    ax.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
    if time_step > 100000:
        ax.ticklabel_format(style='sci', axis='x', scilimits=(0, 0))

    # Output plot
    if show_plot:
        plt.show()
    else:
        fign = dirname +'/RF_freq.png'
        plt.savefig(fign)
    plt.clf()     


def plot_PL_freq_corr(RFStation, h5data, output_freq = 1, 
                      dirname = 'fig', show_plot = False):
    
    """
    Plot of frequency correction applied by the Phase Loop as a function of time.
    For large amount of data, monitor with larger 'output_freq'.
    """

    # Time step of plotting
    time_step = RFStation.counter[0]

    # Load/create data
    if output_freq < 1:
        output_freq = 1
    ndata = int(time_step / output_freq)
    t = output_freq * np.arange(ndata)
    dphi = np.array(h5data["/Beam/PL_omegaRF_corr"][0:ndata], dtype=np.double)
    dphi[time_step:] = np.nan

    # Plot
    fig = plt.figure(1)
    fig.set_size_inches(8,6)
    ax = plt.axes()
    ax.plot(t, dphi,'.')
    ax.set_xlabel(r"No. turns [T$_0$]")    
    ax.set_ylabel (r"PL $\omega_{\mathsf{RF}}$ correction [1/s]")
    ax.ticklabel_format(style='sci', axis='y', scilimits=(0,0))
    if time_step > 100000:
        ax.ticklabel_format(style='sci', axis='x', scilimits=(0, 0))

    # Output plot
    if show_plot:
        plt.show()
    else:
        fign = dirname +'/PL_freq_corr.png'
        plt.savefig(fign)
    plt.clf()     


def plot_RF_phase_error(RFStation, h5data, output_freq = 1, 
                       dirname = 'fig', show_plot = False):
    
    """
    Plot of accumulated RF phase error; the Synchro Loop can act on this.
    For large amount of data, monitor with larger 'output_freq'.
    """

    # Time step of plotting
    time_step = RFStation.counter[0]

    # Load/create data
    if output_freq < 1:
        output_freq = 1
    ndata = int(time_step / output_freq)
    t = output_freq * np.arange(ndata)
    dphi = np.array(h5data["/Beam/SL_dphiRF"][0:ndata], dtype=np.double)
    dphi[time_step:] = np.nan

    # Plot
    plt.figure(1, figsize=(8,6))
    ax = plt.axes()
    ax.plot(t, dphi, '.')
    ax.set_xlabel(r"No. turns [T$_0$]")
    ax.set_ylabel(r"RF phase error $\Delta \phi_{\mathsf{RF}}$ [rad]")
    ax.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
    if time_step > 100000:
        ax.ticklabel_format(style='sci', axis='x', scilimits=(0, 0))

    # Output plot
    if show_plot:
        plt.show()
    else:
        fign = dirname +'/RF_phase_error.png'
        plt.savefig(fign)
    plt.clf()    
    
    
     
def plot_RL_radial_error(RFStation, h5data, output_freq = 1, 
                         dirname = 'fig', show_plot = False):
    
    """
    Plot of relative radial error; monitored with Phase Loop.
    For large amount of data, monitor with larger 'output_freq'.
    """

    # Time step of plotting
    time_step = RFStation.counter[0]

    # Load/create data
    if output_freq < 1:
        output_freq = 1
    ndata = int(time_step / output_freq)
    t = output_freq * np.arange(ndata)
    dphi = np.array(h5data["/Beam/RL_drho"][0:ndata], dtype=np.double)
    dphi[time_step:] = np.nan

    # Plot
    plt.figure(1, figsize=(8,6))
    ax = plt.axes()
    ax.plot(t, dphi, '.')
    ax.set_xlabel(r"No. turns [T$_0$]")
    ax.set_ylabel(r"Relative radial error [1]")
    ax.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
    if time_step > 100000:
        ax.ticklabel_format(style='sci', axis='x', scilimits=(0, 0))

    # Output plot
    if show_plot:
        plt.show()
    else:
        fign = dirname +'/RL_radial_error.png'
        plt.savefig(fign)
    plt.clf()     
    
               

def plot_COM_motion(Ring, RFStation, h5data,  
                    output_freq = 1, dirname = 'fig', show_plot = False):
    """
    Evolution of bunch C.O.M. in longitudinal phase space. 
    Optional use of histograms and separatrix.
    """

    # Time step of plotting
    time_step = RFStation.counter[0]

    # Load/create data
    if output_freq < 1:
        output_freq = 1
    ndata = int(time_step / output_freq)

    # Load data
    mean_dt = np.array(h5data["/Beam/mean_dt"][0:ndata], dtype=np.double)
    mean_dE = np.array(h5data["/Beam/mean_dE"][0:ndata], dtype=np.double)

    # Plot
    fig = plt.figure(1)
    fig.set_size_inches(8,8)
    ax = plt.axes()
    ax.scatter(mean_dt, mean_dE, s=5, edgecolor='none')

    ax.set_xlabel(r"$\Delta t$ [s]")
    ax.set_ylabel(r"$\Delta$E [eV]")
    ax.ticklabel_format(style='sci', axis='x', scilimits=(0, 0))
    ax.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
    ax.set_xlim((-0.7e-6, 0.7e-6))
    plt.figtext(0.95, 0.95, 'C.O.M. evolution', fontsize=16, ha='right', 
                va='center') 
            
    # Output plot
    if show_plot:
        plt.show()
    else:
        fign = dirname +'/COM_evolution.png'
        plt.savefig(fign)
    plt.clf()


def plot_LHCNoiseFB(RFStation, LHCNoiseFB, h5data, output_freq = 1, 
                    dirname = 'fig', show_plot = False):
    
    """
    Plot of the phase noise multiplication factor as a function of time.
    For large amount of data, monitor with larger 'output_freq'.
    """

    # Time step of plotting
    time_step = RFStation.counter[0]

    # Load/create data
    if output_freq < 1:
        output_freq = 1
    ndata = int(time_step / output_freq)
    t = output_freq * np.arange(ndata)
    x = np.array(h5data["/Beam/LHC_noise_FB_factor"][0:ndata], dtype=np.double)
    x[time_step:] = np.nan

    # Plot
    fig = plt.figure(1)
    fig.set_size_inches(8,6)
    ax = plt.axes()
    ax.plot(t, x,'.')
    ax.set_xlabel(r"No. turns [T$_0$]")    
    ax.set_ylabel (r"LHC noise FB scaling factor [1]")
    ax.ticklabel_format(style='sci', axis='y', scilimits=(0,0))
    if time_step > 100000:
        ax.ticklabel_format(style='sci', axis='x', scilimits=(0, 0))

    # Output plot
    if show_plot:
        plt.show()
    else:
        fign = dirname +'/LHC_noise_FB.png'
        plt.savefig(fign)
    plt.clf()         


    
def plot_LHCNoiseFB_FWHM(RFStation, LHCNoiseFB, h5data, 
                         output_freq = 1, dirname = 'fig', show_plot = False):
    
    """
    Plot of the FWHM bunch length used in LHCNoiseFB as a function of time.
    For large amount of data, monitor with larger 'output_freq'.
    """

    # Time step of plotting
    time_step = RFStation.counter[0]

    # Load/create data
    if output_freq < 1:
        output_freq = 1
    ndata = int(time_step / output_freq)
    t = output_freq * np.arange(ndata)
    x = np.array(h5data["/Beam/LHC_noise_FB_bl"][0:ndata], dtype=np.double)
    x[time_step:] = np.nan

    # Plot
    fig = plt.figure(1)
    fig.set_size_inches(8,6)
    ax = plt.axes()
    ax.plot(t, x,'.')
    ax.set_xlabel(r"No. turns [T$_0$]")    
    ax.set_ylabel (r"4-sigma FWHM bunch length [s]")
    if time_step > 100000:
        ax.ticklabel_format(style='sci', axis='x', scilimits=(0, 0))

    # Output plot
    if show_plot:
        plt.show()
    else:
        fign = dirname +'/LHC_noise_FB_bl.png'
        plt.savefig(fign)
    plt.clf()         



def plot_LHCNoiseFB_FWHM_bbb(RFStation, LHCNoiseFB, h5data, 
                             output_freq = 1, dirname = 'fig', show_plot = False):
    
    """
    Plot of bunch-by-bunch FWHM bunch length used in LHCNoiseFB as a function 
    of time. For large amount of data, monitor with larger 'output_freq'.
    """

    # Time step of plotting
    time_step = RFStation.counter[0]

    # Load/create data
    if output_freq < 1:
        output_freq = 1
    ndata = int(time_step / output_freq)
    t = output_freq * np.arange(ndata)
    x = np.array(h5data["/Beam/LHC_noise_FB_bl_bbb"][0:ndata, :], ndmin=2)
    x[time_step:, :] = np.nan
    nbunches = x.shape[1]

    # Plot
    fig = plt.figure(1)
    fig.set_size_inches(8,6)
    ax = plt.axes()
    for i in range(nbunches):
        ax.plot(t, x[:, i], '.', color=cm.get_cmap('jet')(i / nbunches), label="Bunch %d" % i)
    ax.set_xlabel(r"No. turns [T$_0$]")
    ax.set_ylabel(r"4-sigma FWHM bunch length [s]")
    if time_step > 100000:
        ax.ticklabel_format(style='sci', axis='x', scilimits=(0, 0))
    ax.legend()

    # Output plot
    if show_plot:
        plt.show()
    else:
        fign = dirname +'/LHC_noise_FB_bl_bbb.png'
        plt.savefig(fign)
    plt.clf()         



# coding: utf8
# Copyright 2014-2020 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

"""
**Filters and methods for control loops**

:Authors: **Birk Emil Karlsen-Bæck**, **Helga Timko**
"""

from __future__ import annotations

# Set up logging
import logging

import numpy as np
from scipy.constants import c
from scipy.sparse import diags
from scipy.sparse.linalg import spsolve

logger = logging.getLogger(__name__)


def cavity_response_sparse_matrix(
    I_beam,
    I_gen,
    n_samples,
    V_ant_init,
    I_gen_init,
    samples_per_rf,
    R_over_Q,
    Q_L,
    detuning,
):
    """Solving the ACS cavity response model as a sparse matrix problem
    for a given set of initial conditions, resonator parameters and
    generator and RF beam currents.

    Parameters
    ----------
    I_beam : complex array
        RF beam current
    I_gen : complex array
        Generator current
    n_samples : int
        Number of samples of the result array - 1
    V_ant_init : complex float
        Initial condition for the antenna voltage
    I_gen_init : complex float
        Initial condition of the generator current, i.e. one sample before the I_gen array
    samples_per_rf : int
        Number of samples per RF period
    R_over_Q : float
        The R over Q of the cavity
    Q_L : float
        The loaded quality factor of the cavity
    detuning : float
        The detuning of the cavity in frequency divided by the rf frequency

    Returns
    -------
    complex array
        The antenna voltage evaluated for the same period as I_beam and I_gen of length n_samples + 1

    """

    # Add a zero at the start of RF beam current
    if len(I_beam) != n_samples + 1:
        I_beam = np.concatenate((np.zeros(1, dtype=complex), I_beam))

    # Check length of the generator current array
    if len(I_gen) != n_samples + 1:
        I_gen = np.concatenate((I_gen_init * np.ones(1, dtype=complex), I_gen))

    # Compute matrix elements
    A = 0.5 * R_over_Q * samples_per_rf
    B = 1 - 0.5 * samples_per_rf / Q_L + 1j * detuning * samples_per_rf

    # Initialize the two sparse matrices needed to find antenna voltage
    B_matrix = diags(
        [-B, 1], [-1, 0], (n_samples + 1, n_samples + 1), dtype=complex, format="csc"
    )
    I_matrix = diags([A], [-1], (n_samples + 1, n_samples + 1), dtype=complex)

    # Find vector on the "current" side of the equation
    b = I_matrix.dot(2 * I_gen - I_beam)
    b[0] = V_ant_init

    # Solve the sparse linear system of equations and return
    return spsolve(B_matrix, b)


def rectangle(t, tau):
    r"""Rectangular function of time

    .. math:: \mathsf{rect} \left( \frac{t}{\tau} \right) =
        \begin{cases}
            1 \, , \, t \in (-\tau/2, \tau/2) \\
            0.5 \, , \, t = \pm \tau/2 \\
            0 \, , \, \textsf{otherwise}
        \end{cases}

    Parameters
    ----------
    t : float array
        Time array
    tau : float
        Time window of rectangular function

    Returns
    -------
    float array
        Rectangular function for given time array

    """

    dt = t[1] - t[0]
    llimit = np.where(np.fabs(t + tau / 2) < dt / 2)[0]
    ulimit = np.where(np.fabs(t - tau / 2) < dt / 2)[0]
    if len(llimit) != 1:
        # ImpulseError
        raise RuntimeError(
            "ERROR in impulse_response.rectangle(): time"
            + " array doesn't start at rising edge!"
        )
    if len(ulimit) not in [0, 1]:
        # ImpulseError
        raise RuntimeError(
            "ERROR in impulse_response.rectangle(): time"
            + " array has multiple falling edges!"
        )
    logger.debug("In rectangle(), index of rising edge is %d" % llimit[0])
    y = np.zeros(len(t))
    y[llimit[0]] = 0.5
    if len(ulimit) == 1:
        y[llimit[0] + 1 : ulimit[0]] = np.ones(ulimit[0] - llimit[0] - 1)
        y[ulimit[0]] = 0.5
    else:
        y[llimit[0] + 1 :] = 1

    return y


def triangle(t, tau):
    r"""Triangular function of time

    .. math:: \mathsf{tri} \left( \frac{t}{\tau} \right) =
        \begin{cases}
            1 - \frac{t}{\tau}\, , \, t \in (0, \tau) \\
            0.5 \, , \, t = 0 \\
            0 \, , \, \textsf{otherwise}
        \end{cases}

    Parameters
    ----------
    t : float array
        Time array
    tau : float
        Time window of rectangular function

    Returns
    -------
    float array
        Triangular function for given time array

    """

    dt = t[1] - t[0]
    llimit = np.where(np.fabs(t) < dt / 2)[0]
    logger.debug("In triangle(), index of rising edge is %d" % llimit[0])
    if len(llimit) != 1:
        # ImpulseError
        raise RuntimeError(
            "ERROR in impulse_response.triangle(): time"
            + " array doesn't start at rising edge!"
        )
    y = np.zeros(len(t))
    y[llimit[0]] = 0.5
    y[llimit[0] + 1 :] = 1 - t[llimit[0] + 1 :] / tau
    y[np.where(y < 0)[0]] = 0

    return y


class TravellingWaveCavity:
    r"""Impulse responses of a travelling wave cavity. The induced voltage
    :math:`V(t)` from the impulse response :math:`h(t)` and the I,Q (cavity or
    generator) current :math:`I(t)` can be written in matrix form,

    .. math::
        \left( \begin{matrix} V_I(t) \\
        V_Q(t) \end{matrix} \right)
        = \left( \begin{matrix} h_s(t) & - h_c(t) \\
        h_c(t) & h_s(t) \end{matrix} \right)
        * \left( \begin{matrix} I_I(t) \\
        I_Q(t) \end{matrix} \right) \, ,

    where :math:`*` denotes convolution,
    :math:`h(t)*x(t) = \int d\tau h(\tau)x(t-\tau)`.

    For the **cavity-to-beam induced voltage**, we define

    .. math::
        R_b \equiv \frac{\rho l^2}{8} \,

    where :math:`\rho` is the series impedance, :math:`l` the accelerating
    length, :math:`\tau` the filling time. The cavity-to-beam wake is

    .. math::
        W_b(t) = \frac{4 R_b}{\tau} \mathsf{tri}\left(\frac{t}{\tau}\right)
         \cos(\omega_r t)

    and the impulse response components are

    .. math::
        h_{s,b}(t) &= \frac{2 R_b}{\tau} \mathsf{tri}\left(\frac{t}{\tau}\right)
         \cos((\omega_c - \omega_r)t) \, , \\
        h_{c,b}(t) &= \frac{2 R_b}{\tau} \mathsf{tri}\left(\frac{t}{\tau}\right)
        \sin((\omega_c - \omega_r)t) \, ,

    where :math:`\mathsf{tri}(x)` is the triangular function, :math:`\omega_r`
    is the central revolution frequency of the cavity, and :math:`\omega_c` is
    the carrier revolution frequency of the I,Q demodulated current signal. On
    the carrier frequency, :math:`\omega_c = \omega_r`,

    .. math::
        h_{s,b}(t) &= \frac{2 R_b}{\tau} \mathsf{tri}\left(\frac{t}{\tau}\right) \\
        h_{c,b}(t) &= 0 \, .

    For the **cavity-to-generator induced voltage**, we define

    .. math::
        R_g \equiv l \sqrt{\frac{\rho Z_0}{2}} \,

    where :math:`Z_0` is the shunt impedance when measuring the generator
    current; assumed to be 50 :math:`\Omega`. The cavity-to-generator wake is

    .. math::
        W_g(t) = \frac{2 R_g}{\tau} \mathsf{rect}\left(\frac{t}{\tau}\right)
        \cos(\omega_r t)

    and the impulse response components are

    .. math::
        h_{s,g}(t) &= \frac{R_g}{\tau} \mathsf{rect}\left(\frac{t}{\tau}\right)
        \cos((\omega_c - \omega_r)t) \, , \\
        h_{c,g}(t) &= \frac{R_g}{\tau} \mathsf{rect}\left(\frac{t}{\tau}\right)
        \sin((\omega_c - \omega_r)t) \, ,

    where :math:`\mathsf{rect}(x)` is the rectangular function. On the carrier
    frequency, :math:`\omega_c = \omega_r`,

    .. math::
        h_{s,g}(t) &= \frac{R_g}{\tau} \mathsf{rect}\left(\frac{t}{\tau}\right) \\
        h_{c,g}(t) &= 0 \, .

    Parameters
    ----------
    l_cell : float
        Cavity cell length [m]
    N_cells : int
        Number of accelerating (interacting) cells in a cavity
    rho : float
        Series impedance [Ohms/m^2] of the cavity
    v_g : float
        Group velocity [c] in units of the speed of light
    omega_r : flaot
        Central (resonance) revolution frequency [1/s] of the cavity

    Attributes
    ----------
    Z_0 : float
        Shunt impedance of generator current measurement; assumed to be 50 Ohms
    R_beam : float
        :math:`R_b` [\Omega] as defined above
    R_gen : float
        :math:`R_g` [\Omega] as defined above
    l_cav : float
        Length [m] of the interaction region
    tau : float
        Cavity filling time [s]

    """

    def __init__(
        self,
        l_cell: float,
        N_cells: int,
        rho: float,
        v_g: float,
        omega_r: float,
        df: float = 0,
    ):
        self.l_cell = float(l_cell)
        self.N_cells = int(N_cells)
        self.rho = float(rho)
        if v_g > 0 and v_g < 1:
            self.v_g = float(v_g)
        else:
            # ImpulseError
            raise RuntimeError(
                "ERROR in TravellingWaveCavity: group"
                + " velocity out of limits (0,1)!"
            )
        self.omega_r = float(omega_r) + 2 * np.pi * float(df)

        # Calculated
        self.l_cav = float(self.l_cell * self.N_cells)
        # v_g opposite to wave!
        self.tau = self.l_cav / (self.v_g * c) * (1 + self.v_g)

        # Assumed impedance for measurement of generator current
        self.Z_0 = 50
        # Shunt impedances towards beam and generator
        self.R_beam = 0.125 * self.rho * self.l_cav**2
        self.R_gen = self.l_cav * np.sqrt(0.5 * self.rho * self.Z_0)

        # Set up logging
        self.logger = logging.getLogger(__class__.__name__)
        self.logger.info("Class initialized")
        self.logger.debug("Filling time %.4e s", self.tau)

    def impulse_response_gen(self, omega_c: float, time_coarse: float):
        r"""Impulse response from the cavity towards the
        generator. For a signal that is I,Q demodulated at a given carrier
        frequency :math:`\omega_c`. The formulae assume that the carrier
        frequency is be close to the central frequency
        :math:`\omega_c/\omega_r \ll 1` and that the signal is low-pass
        filtered (i.e.\ high-frequency components can be neglected).

        Parameters
        ----------
        omega_c : float
            Carrier revolution frequency [1/s]
        time_coarse : float
            Time array of the LLRF to act on

        Attributes
        ----------
        d_omega : float
            :math:`\omega_c - \omega_r` [1/s]
        t_gen : float array
            time array for generator wake and impulse response; starts from
            :math:`- \tau/2`
        h_gen : complex array
            :math:`h_{s,b}(t) + i*h_{c,b}(t)` [\Omega/s] as defined above
        """

        self.omega_c = float(omega_c)
        self.d_omega = self.omega_c - self.omega_r
        if np.fabs((self.d_omega) / self.omega_r) > 0.1:
            # ImpulseError
            raise RuntimeError(
                "ERROR in TravellingWaveCavity"
                + " impulse_response(): carrier frequency"
                + " should be close to central frequency of the"
                + " cavity!"
            )

        # Move starting point of impulse response to correct value
        t_gen = time_coarse - time_coarse[0]

        # Impulse response if on carrier frequency
        self.h_gen = (
            self.R_gen / self.tau * rectangle(t_gen - 0.5 * self.tau, self.tau)
        ).astype(np.complex128)

        # Impulse response if not on carrier frequency
        if np.fabs((self.d_omega) / self.omega_r) > 1e-12:
            self.h_gen = self.h_gen.real * (
                np.cos(self.d_omega * t_gen) - 1j * np.sin(self.d_omega * t_gen)
            )

    def impulse_response_beam(
        self, omega_c: float, time_fine: float, time_coarse: float = None
    ):
        r"""Impulse response from the cavity towards the beam. For a signal
        that is I,Q demodulated at a given carrier
        frequency :math:`\omega_c`. The formulae assume that the carrier
        frequency is be close to the central frequency
        :math:`\omega_c/\omega_r \ll 1` and that the signal is low-pass
        filtered (i.e.\ high-frequency components can be neglected).

        Parameters
        ----------
        omega_c : float
            Carrier revolution frequency [1/s]
        time_fine : float
            Time array of the beam profile to act on
        time_coarse : float
            Time array of the LLRF to act on; default is None

        Attributes
        ----------
        d_omega : float
            :math:`\omega_c - \omega_r` [1/s]
        t_beam : float array
            time array for beam wake and impulse response; starts from zero
        h_beam : complex array
            :math:`h_{s,b}(t) + i*h_{c,b}(t)` [\Omega/s] as defined above
        h_beam_coarse : complex array
            Impulse response evaluated on the coarse grid
        """

        self.omega_c = float(omega_c)
        self.d_omega = self.omega_c - self.omega_r
        if np.fabs((self.d_omega) / self.omega_r) > 0.1:
            raise RuntimeError(
                "ERROR in TravellingWaveCavity"
                + " impulse_response(): carrier frequency"
                + " should be close to central frequency of the"
                + " cavity!"
            )

        # Move starting point of impulse response to correct value
        t_beam = time_fine - time_fine[0]

        # Impulse response if on carrier frequency
        self.h_beam = (-2 * self.R_beam / self.tau * triangle(t_beam, self.tau)).astype(
            np.complex128
        )

        # Impulse response if not on carrier frequency
        if np.fabs((self.d_omega) / self.omega_r) > 1e-12:
            self.h_beam = self.h_beam.real * (
                np.cos(self.d_omega * t_beam) - 1j * np.sin(self.d_omega * t_beam)
            )

        if time_coarse is not None:
            # Move starting point of impulse response to correct value
            t_beam = time_coarse - time_coarse[0]

            # Impulse response if on carrier frequency
            self.h_beam_coarse = (
                -2 * self.R_beam / self.tau * triangle(t_beam, self.tau)
            ).astype(np.complex128)

            # Impulse response if not on carrier frequency
            if np.fabs((self.d_omega) / self.omega_r) > 1e-12:
                self.h_beam_coarse = self.h_beam_coarse.real * (
                    np.cos(self.d_omega * t_beam) - 1j * np.sin(self.d_omega * t_beam)
                )

    def compute_wakes(self, time: float):
        r"""Computes the wake fields towards the beam and generator on the
        central cavity frequency.

        Parameters
        ----------
        time_beam : float
            Time array of the beam to act on
        time_gen : float
            Time array of the generator to act on

        Attributes
        ----------
        W_beam : float array
            :math:`W_b(t)` [\Omega/s] as defined above
        W_gen : float array
            :math:`W_g(t)` [\Omega/s] as defined above

        """

        t_beam = time - time[0]
        t_gen = time - time[0] - 0.5 * self.tau

        # Wake fields towards beam and generator
        self.W_beam = 2 * self.h_beam.real * np.cos(self.omega_r * t_beam)
        self.W_gen = 2 * self.h_gen.real * np.cos(self.omega_r * t_gen)


class SPS3Section200MHzTWC(TravellingWaveCavity):
    def __init__(self, df: float = 0):
        TravellingWaveCavity.__init__(
            self, 0.374, 32, 2.71e4, 0.0946, 2 * np.pi * 200.03766667e6, df=df
        )


class SPS4Section200MHzTWC(TravellingWaveCavity):
    def __init__(self, df: float = 0):
        TravellingWaveCavity.__init__(
            self, 0.374, 43, 2.71e4, 0.0946, 2 * np.pi * 199.9945e6, df=df
        )


class SPS5Section200MHzTWC(TravellingWaveCavity):
    def __init__(self, df: float = 0):
        TravellingWaveCavity.__init__(
            self, 0.374, 54, 2.71e4, 0.0946, 2 * np.pi * 200.1e6, df=df
        )




# ===============
# Beam Exceptions
# ===============

class MassError(Exception):
    pass


class AllParticlesLost(Exception):
    pass


class ParticleAdditionError(Exception):
    pass


# ==================================
# Distribution Generation Exceptions
# ==================================

class DistributionError(Exception):
    pass


class GenerationError(Exception):
    pass


# ==================
# Profile Exceptions
# ==================

class CutError(Exception):
    pass


class ProfileDerivativeError(Exception):
    pass


# ====================
# Impedance Exceptions
# ====================

class WakeLengthError(Exception):
    pass


class FrequencyResolutionError(Exception):
    pass


class ResonatorError(Exception):
    pass


class WrongCalcError(Exception):
    pass


class MissingParameterError(Exception):
    pass


# ===========================
# Input Parameters Exceptions
# ===========================

class MomentumError(Exception):
    pass


# ===============
# LLRF Exceptions
# ===============

class PhaseLoopError(Exception):
    pass


class PhaseNoiseError(Exception):
    pass


class FeedbackError(Exception):
    pass


class ImpulseError(Exception):
    pass


# ==================
# Toolbox Exceptions
# ==================

class PhaseSpaceError(Exception):
    pass


class NoiseDiffusionError(Exception):
    pass


# ==================
# Tracker Exceptions
# ==================

class PotentialWellError(Exception):
    pass


class SolverError(Exception):
    pass


class PeriodicityError(Exception):
    pass


class ProfileError(Exception):
    pass


class SynchrotronMotionError(Exception):
    pass


# ===============
# Util Exceptions
# ===============

class ConvolutionError(Exception):
    pass


class IntegrationError(Exception):
    pass


class SortError(Exception):
    pass


# =================
# Global Exceptions
# =================

class InterpolationError(Exception):
    pass


class InputDataError(Exception):
    pass




# Copyright 2016 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
**Class to compute synchrotron radiation damping and quantum excitation**

:Authors: **Juan F. Esteban Mueller**
'''

from __future__ import division, print_function

from builtins import range

import numpy as np

from ..utils import bmath as bm


class SynchrotronRadiation:

    ''' Class to compute synchrotron radiation effects, including radiation
        damping and quantum excitation.
        For multiple RF section, instanciate one object per RF section an call
        the track() method after tracking each section.
    '''

    def __init__(self, Ring, RFParameters, Beam, bending_radius,
                 n_kicks=1, quantum_excitation=True, python=False, seed=None,
                 shift_beam=True):

        self.ring = Ring
        self.rf_params = RFParameters
        self.beam = Beam
        self.rho = bending_radius
        self.n_kicks = n_kicks  # To apply SR in several kicks
        np.random.seed(seed=seed)

        # Calculate static parameters
        self.c_gamma = self.ring.Particle.c_gamma
        self.c_q = self.ring.Particle.c_q

        self.I2 = 2.0 * np.pi / self.rho     # Assuming isomagnetic machine
        self.I3 = 2.0 * np.pi / self.rho**2.0
        self.I4 = self.ring.ring_circumference * self.ring.alpha_0[0, 0] / self.rho**2.0
        self.jz = 2.0 + self.I4 / self.I2

        # Calculate synchrotron radiation parameters
        self.calculate_SR_params()

        # Initialize the random number array if quantum excitation is included
        if quantum_excitation:
            self.random_array = np.zeros(self.beam.n_macroparticles)

        # Displace the beam in phase to account for the energy loss due to
        # synchrotron radiation (temporary until bunch generation is updated)
        if (shift_beam) and (self.rf_params.section_index == 0):
            self.beam_phase_to_compensate_SR = np.abs(np.arcsin(
                self.U0 / (self.ring.Particle.charge * self.rf_params.voltage[0][0])))
            self.beam_position_to_compensate_SR = self.beam_phase_to_compensate_SR \
                * self.rf_params.t_rf[0, 0] / (2.0 * np.pi)

            self.beam.dt -= self.beam_position_to_compensate_SR

        # Select the right method for the tracker according to the selected
        # settings
        if python:
            if quantum_excitation:
                self.track = self.track_full_python
            else:
                self.track = self.track_SR_python
        else:
            if quantum_excitation:
                if seed is not None:
                    bm.set_random_seed(seed)
                self.track = self.track_full_C
            else:
                self.track = self.track_SR_C

    # Method to compute the SR parameters
    def calculate_SR_params(self):
        i_turn = self.rf_params.counter[0]

        # Energy loss per turn/RF section [eV]
        self.U0 = (self.c_gamma * self.ring.energy[0, i_turn]**4.0
                   * self.I2 / (2.0 * np. pi)
                   * self.rf_params.section_length
                   / self.ring.ring_circumference)

        # Damping time [turns]
        self.tau_z = (2.0 / self.jz * self.ring.energy[0, i_turn] /
                      self.U0)

        # Equilibrium energy spread
        self.sigma_dE = np.sqrt(self.c_q *
                                self.ring.gamma[0, i_turn]**2.0 *
                                self.I3 / (self.jz * self.I2))

    # Print SR parameters
    def print_SR_params(self):
        i_turn = self.rf_params.counter[0]

        print('------- Synchrotron radiation parameters -------')
        print(f'jz = {self.jz:1.8f}')
        if (self.rf_params.section_length
                == self.ring.ring_circumference):
            print(f'Energy loss per turn = {self.U0/1e9:1.4f} GeV/turn')
            print(f'Damping time = {self.tau_z:1.4f} turns')
        else:
            print('Energy loss per RF section = {0:1.4f} GeV/section'.format(
                self.U0 * 1e-9))
            print('Energy loss per turn = {0:1.4f} GeV/turn'.format(
                self.U0 * 1e-9 * self.ring.ring_circumference
                / self.rf_params.section_length))
            print('Damping time = {0:1.4f} turns'.format(self.tau_z
                                                         * self.rf_params.section_length
                                                         / self.ring.ring_circumference))
        print(f'Equilibrium energy spread = {self.sigma_dE * 100:1.4f}%'
              + f' ({self.sigma_dE * self.ring.energy[0, i_turn]*1e-6:1.4f}) MeV')
        print('------------------------------------------------')

    # Track particles with SR only (without quantum excitation)
    def track_SR_python(self):
        i_turn = self.rf_params.counter[0]
        # Recalculate SR parameters if energy changes
        if (i_turn != 0 and self.ring.energy[0, i_turn] !=
                self.ring.energy[0, i_turn - 1]):
            self.calculate_SR_params()
        for i in range(self.n_kicks):
            self.beam.dE += -(2.0 / self.tau_z / self.n_kicks * self.beam.dE
                              + self.U0 / self.n_kicks)

    # Track particles with SR and quantum excitation
    def track_full_python(self):
        i_turn = self.rf_params.counter[0]
        # Recalculate SR parameters if energy changes
        if (i_turn != 0 and self.ring.energy[0, i_turn] !=
                self.ring.energy[0, i_turn - 1]):
            self.calculate_SR_params()
        for i in range(self.n_kicks):
            self.beam.dE += -(2.0 / self.tau_z / self.n_kicks * self.beam.dE +
                              self.U0 / self.n_kicks - 2.0 * self.sigma_dE /
                              np.sqrt(self.tau_z * self.n_kicks) *
                              self.ring.energy[0, i_turn] *
                              np.random.randn(self.beam.n_macroparticles))

    # Track particles with SR only (without quantum excitation)
    # C implementation
    def track_SR_C(self):
        i_turn = self.rf_params.counter[0]
        # Recalculate SR parameters if energy changes
        if (i_turn != 0 and self.ring.energy[0, i_turn] !=
                self.ring.energy[0, i_turn - 1]):
            self.calculate_SR_params()

        bm.synchrotron_radiation(self.beam.dE, self.U0,
                                 self.n_kicks, self.tau_z)

    # Track particles with SR and quantum excitation. C implementation
    def track_full_C(self):
        i_turn = self.rf_params.counter[0]
        # Recalculate SR parameters if energy changes
        if (i_turn != 0 and self.ring.energy[0, i_turn] !=
                self.ring.energy[0, i_turn - 1]):
            self.calculate_SR_params()

        bm.synchrotron_radiation_full(self.beam.dE, self.U0, self.n_kicks,
                                      self.tau_z, self.sigma_dE,
                                      self.ring.energy[0, i_turn])

    def to_gpu(self, recursive=True):
        '''
        Transfer all necessary arrays to the GPU
        '''
        # Check if to_gpu has been invoked already
        if hasattr(self, '_device') and self._device == 'GPU':
            return

        # No arrays need to be transfered

        # to make sure it will not be called again
        self._device = 'GPU'

    def to_cpu(self, recursive=True):
        '''
        Transfer all necessary arrays back to the CPU
        '''
        # Check if to_cpu has been invoked already
        if hasattr(self, '_device') and self._device == 'CPU':
            return

        # No arrays need to be transfered

        # to make sure it will not be called again
        self._device = 'CPU'




# Copyright 2016 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
**Module to generate coasting beam**

:Authors: **Simon Albright**
'''

from builtins import str

import numpy as np
import numpy.random as rand

import blond.utils.exceptions as blExcept



def generate_coasting_beam(beam, t_start, t_stop, spread=1E-3,
                           spread_type='dp/p', energy_offset=0,
                           distribution='gaussian', user_distribution=None,
                           user_probability=None):
    '''
    energy_offset represents the absolute energy difference between the center
    of the coasting beam and the synchronous particle spread is used for the
    absolute max/min in a parabolic or the standard deviation in a gaussian
    distribution spread_type is used to convert the passed spread into dE
    according to dE/E = beta**2 * dP/P energy_offset gives an offset in dE for
    the two standard distributions if a user_distribution is used it is taken as
    being in dE

    :param beam: _description_
    :type beam: _type_
    :param t_start: _description_
    :type t_start: _type_
    :param t_stop: _description_
    :type t_stop: _type_
    :param spread: _description_, defaults to 1E-3
    :type spread: _type_, optional
    :param spread_type: _description_, defaults to 'dp/p'
    :type spread_type: str, optional
    :param energy_offset: _description_, defaults to 0
    :type energy_offset: int, optional
    :param distribution: _description_, defaults to 'gaussian'
    :type distribution: str, optional
    :param user_distribution: _description_, defaults to None
    :type user_distribution: _type_, optional
    :param user_probability: _description_, defaults to None
    :type user_probability: _type_, optional
    :raises blExcept.DistributionError: _description_
    :raises blExcept.DistributionError: _description_
    :raises blExcept.DistributionError: _description_
    '''
    if spread_type == 'dp/p':
        energy_spread = beam.energy * beam.beta**2 * spread
    elif spread_type == 'dE/E':
        energy_spread = spread * beam.energy
    elif spread_type == 'dp':
        energy_spread = beam.energy * beam.beta**2 * spread / beam.momentum
    elif spread_type == 'dE':
        energy_spread = spread
    else:
        raise blExcept.DistributionError("spread_type " + str(spread_type) +
                                         " not recognised")

    if distribution == 'gaussian':
        beam.dE = rand.normal(loc=energy_offset, scale=energy_spread,
                              size=beam.n_macroparticles)

    elif distribution == 'parabolic':
        energy_range = np.linspace(-energy_spread, energy_spread, 10000)
        probability_distribution = 1 - (energy_range / energy_spread)**2
        probability_distribution /= np.cumsum(probability_distribution)[-1]
        beam.dE = rand.choice(energy_range, size=beam.n_macroparticles,
                              p=probability_distribution) \
            + (rand.rand(beam.n_macroparticles) - 0.5) \
            * (energy_range[1] - energy_range[0]) \
            + energy_offset

    # If distribution == 'user' is selected the user must supply a uniformly
    # spaced distribution and the assosciated probability for each bin
    # momentum_spread and energy_offset are not used in this instance.
    elif distribution == 'user':
        if user_distribution is None or user_probability is None:
            raise blExcept.DistributionError("""Distribution 'user' requires
                                             'user_distribution' and 
                                             'user_probability' to be defined""")

        beam.dE = rand.choice(user_distribution, size=beam.n_macroparticles,
                              p=user_probability) \
            + (rand.rand(beam.n_macroparticles) - 0.5) \
            * (user_distribution[1] - user_distribution[0])

    else:
        raise blExcept.DistributionError("distribution type not recognised")

    beam.dt = rand.rand(beam.n_macroparticles) * (t_stop - t_start) + t_start



'''
Functions related to running MPI simulations.

@author: Konstantinos Iliakis
@date: 01.01.2020
'''

import logging
import os
import socket
import sys
from functools import wraps

import numpy as np
from mpi4py import MPI

from ..utils import bmath as bm

WORKER = None


def mpiprint(*args, all=False):
    """Masks default print function, so that the worker id is also printed

    Args:
        all (bool, optional): _description_. Defaults to False.
    """
    if WORKER.is_master or all:
        print(f'[{WORKER.rank}]', *args)


def master_wrap(func):
    """Wrap function to be executed only by the master worker.

    Args:
        func (_type_): _description_

    Returns:
        _type_: _description_
    """
    @wraps(func)
    def wrap(*args, **kwargs):
        if WORKER.is_master:
            return func(*args, **kwargs)
        return None
    return wrap


def sequential_wrap(func, beam, split_args={}, gather_args={}):
    """Wrap a function to make it run in sequential mode.
    When in sequential mode, all the beam coordinates are gathered before executing
    the passed function, and re-splitted afterwards.

    Args:
        func (_type_): _description_
        beam (_type_): _description_
        split_args (dict, optional): _description_. Defaults to {}.
        gather_args (dict, optional): _description_. Defaults to {}.

    Returns:
        _type_: _description_
    """
    @wraps(func)
    def wrap(*args, **kw):
        beam.gather(**gather_args)
        if WORKER.is_master:
            result = func(*args, **kw)
        else:
            result = None
        beam.split(**split_args)
        return result
    return wrap


class Worker:
    """Stores information accessed by each MPI worker. Also contains all needed MPI methods.
    """
    def __init__(self):
        """Constructor
        """
        self.log = False
        self.logger = None
        # Global inter-communicator
        self.intercomm = MPI.COMM_WORLD
        self.rank = self.intercomm.rank
        self.workers = self.intercomm.size

        # Get hostname
        self.hostname = MPI.Get_processor_name()

        # Get host IP
        self.hostip = socket.gethostbyname(self.hostname)
        # Create communicator with processes on the same host
        color = np.dot(np.array(self.hostip.split('.'), int)
                       [1:], [1, 256, 256**2])
        self.nodecomm = self.intercomm.Split(color, self.rank)
        self.noderank = self.nodecomm.rank
        self.nodeworkers = self.nodecomm.size

        # Assign default values for GPUs
        self.gpu_id = -1
        self.has_gpu = False

    def assign_gpus(self, num_gpus=0):
        """Assign GPUs to workers

        Args:
            num_gpus (int, optional): _description_. Defaults to 0.
        """
        if self.noderank < num_gpus:
            self.has_gpu = True
            self.gpu_id = self.noderank

    def init_log(self, log, logdir):
        """Initialize the logs

        Args:
            log (_type_): _description_
            logdir (_type_): _description_
        """
        self.log = log
        self.logger = MPILog(rank=self.rank, log_dir=logdir)
        if not self.log:
            self.logger.disable()

    @property
    def is_master(self):
        """Return true if this worker is the master

        Returns:
            _type_: _description_
        """
        return self.rank == 0

    # Define the begin and size numbers in order to split a variable of length size
    def gather(self, var):
        """Gather a vector on the master

        Args:
            var (_type_): _description_

        Returns:
            _type_: _description_
        """
        if self.log:
            self.logger.debug('gather')

        # First I need to know the total size
        counts = np.zeros(self.workers, dtype=int)
        sendbuf = np.array([len(var)], dtype=int)
        self.intercomm.Gather(sendbuf, counts, root=0)
        total_size = np.sum(counts)

        if self.is_master:
            # counts = [size // self.workers + 1 if i < size % self.workers
            #           else size // self.workers for i in range(self.workers)]
            displs = np.append([0], np.cumsum(counts[:-1]))
            sendbuf = np.copy(var)
            recvbuf = np.resize(var, total_size)

            self.intercomm.Gatherv(sendbuf,
                                   [recvbuf, counts, displs, recvbuf.dtype.char], root=0)
            return recvbuf
        else:
            recvbuf = None
            self.intercomm.Gatherv(var, recvbuf, root=0)
            return var

    # All workers gather the variable var (from all workers)

    def allgather(self, var):
        """Gather vector to all workers.

        Args:
            var (_type_): _description_

        Returns:
            _type_: _description_
        """
        if self.log:
            self.logger.debug('allgather')

        # One first gather to collect all the sizes
        counts = np.zeros(self.workers, dtype=int)
        sendbuf = np.array([len(var)], dtype=int)
        self.intercomm.Allgather(sendbuf, counts)

        total_size = np.sum(counts)
        # counts = [size // self.workers + 1 if i < size % self.workers
        #           else size // self.workers for i in range(self.workers)]
        displs = np.append([0], np.cumsum(counts[:-1]))
        sendbuf = np.copy(var)
        recvbuf = np.resize(var, total_size)

        self.intercomm.Allgatherv(sendbuf,
                                  [recvbuf, counts, displs, recvbuf.dtype.char])
        return recvbuf

    def scatter(self, var):
        """Scatter vector from master to all workers.

        Args:
            var (_type_): _description_

        Returns:
            _type_: _description_
        """
        if self.log:
            self.logger.debug('scatter')

        # First broadcast the total_size from the master
        total_size = int(self.intercomm.bcast(len(var), root=0))

        # Then calculate the counts (size for each worker)
        counts = [total_size // self.workers + 1 if i < total_size % self.workers
                  else total_size // self.workers for i in range(self.workers)]

        if self.is_master:
            displs = np.append([0], np.cumsum(counts[:-1]))
            recvbuf = np.empty(counts[WORKER.rank], dtype=var.dtype.char)
            self.intercomm.Scatterv([var, counts, displs, var.dtype.char],
                                    recvbuf, root=0)
        else:
            sendbuf = None
            recvbuf = np.empty(counts[WORKER.rank], dtype=var.dtype.char)
            self.intercomm.Scatterv(sendbuf, recvbuf, root=0)

        return recvbuf

    def broadcast(self, var, root=0):
        """Broadcast array to all workers.

        Args:
            var (_type_): _description_
            root (int, optional): _description_. Defaults to 0.

        Returns:
            _type_: _description_
        """
        if self.log:
            self.logger.debug('broadcast')

        # First broadcast the size and dtype from the master
        # recvbuf = self.intercomm.bcast([len(var), var.dtype.char], root=0)
        # size, dtype = recvbuf[0], recvbuf[1]

        if self.is_master:
            recvbuf = self.intercomm.bcast(var, root=root)
        else:
            recvbuf = None
            recvbuf = self.intercomm.bcast(recvbuf, root=root)

        return recvbuf

    def reduce(self, sendbuf, recvbuf=None, dtype=np.uint32, operator='custom_sum',
               comm=None):
        """Reduce array to master.

        Args:
            sendbuf (_type_): _description_
            recvbuf (_type_, optional): _description_. Defaults to None.
            dtype (_type_, optional): _description_. Defaults to np.uint32.
            operator (str, optional): _description_. Defaults to 'custom_sum'.
            comm (_type_, optional): _description_. Defaults to None.

        Returns:
            _type_: _description_
        """
        if comm is None:
            comm = self.intercomm
        # supported ops:
        # sum, mean, std, max, min, prod, custom_sum
        if self.log:
            self.logger.debug('reduce')
        operator = operator.lower()
        if operator in ['sum', 'custom_sum']:
            mpi_op = MPI.SUM
        elif operator == 'max':
            mpi_op = MPI.MAX
        elif operator == 'min':
            mpi_op = MPI.MIN
        elif operator == 'prod':
            mpi_op = MPI.PROD
        elif operator in ['mean', 'avg']:
            mpi_op = MPI.SUM
        elif operator == 'std':
            recvbuf = self.gather(sendbuf)
            if WORKER.is_master:
                assert len(recvbuf) == 3 * self.workers
                totals = np.sum((recvbuf[2::3] - 1) * recvbuf[1::3]**2 +
                                recvbuf[2::3] * (recvbuf[1::3] - bm.mean(recvbuf[0::3]))**2)
                return np.array([np.sqrt(totals / (np.sum(recvbuf[2::3]) - 1))])
            else:
                return np.array([sendbuf[1]])

        if WORKER.is_master:
            if (recvbuf is None) or (sendbuf is recvbuf):
                comm.Reduce(MPI.IN_PLACE, sendbuf, op=mpi_op, root=0)
                recvbuf = sendbuf
            else:
                comm.Reduce(sendbuf, recvbuf, op=mpi_op, root=0)

            if operator in ['mean', 'avg']:
                return recvbuf / self.workers
            return recvbuf
        else:
            recvbuf = None
            comm.Reduce(sendbuf, recvbuf, op=mpi_op, root=0)
            return sendbuf

    def allreduce(self, sendbuf, recvbuf=None, dtype=np.uint32, operator='sum',
                  comm=None):
        """Reduce array to all workers.

        Args:
            sendbuf (_type_): _description_
            recvbuf (_type_, optional): _description_. Defaults to None.
            dtype (_type_, optional): _description_. Defaults to np.uint32.
            operator (str, optional): _description_. Defaults to 'sum'.
            comm (_type_, optional): _description_. Defaults to None.

        Returns:
            _type_: _description_
        """
        if comm is None:
            comm = self.intercomm

        # supported ops:
        # sum, mean, std, max, min, prod, custom_sum
        if self.log:
            self.logger.debug('allreduce')
        operator = operator.lower()

        if operator in ['sum', 'custom_sum']:
            mpi_op = MPI.SUM
        elif operator == 'max':
            mpi_op = MPI.MAX
        elif operator == 'min':
            mpi_op = MPI.MIN
        elif operator == 'prod':
            mpi_op = MPI.PROD
        elif operator in ['mean', 'avg']:
            mpi_op = MPI.SUM
        elif operator == 'std':
            recvbuf = self.allgather(sendbuf)
            assert len(recvbuf) == 3 * self.workers
            totals = np.sum((recvbuf[2::3] - 1) * recvbuf[1::3]**2 +
                            recvbuf[2::3] * (recvbuf[1::3] - bm.mean(recvbuf[::3]))**2)
            return np.array([np.sqrt(totals / (np.sum(recvbuf[2::3]) - 1))])

        if (recvbuf is None) or (sendbuf is recvbuf):
            comm.Allreduce(MPI.IN_PLACE, sendbuf, op=mpi_op)
            recvbuf = sendbuf
        else:
            comm.Allreduce(sendbuf, recvbuf, op=mpi_op)

        if operator in ['mean', 'avg']:
            return recvbuf / self.workers
        return recvbuf

    def sync(self):
        """Synchronize all workers.
        """
        if self.log:
            self.logger.debug('sync')
        self.intercomm.Barrier()

    def finalize(self):
        """Leave MPI.
        """
        if self.log:
            self.logger.debug('finalize')
        if not self.is_master:
            sys.exit(0)

    def greet(self):
        """Print greeting message
        """
        if self.log:
            self.logger.debug('greet')
        print(f'[{self.rank}]@{self.hostname}: Hello World!')

    def print_version(self):
        """Print MPI version.
        """
        if self.log:
            self.logger.debug('version')
        print(f'[{self.rank}] Library: {MPI.get_vendor()}')


class MPILog:
    """Class to log messages coming from other classes. Messages contain
    {Time stamp} {Class name} {Log level} {Message}. Errors, warnings and info
    are logged into the console. To disable logging, call Logger().disable()
    Parameters
    ----------
    debug : bool
        Log DEBUG messages in 'debug.log'; default is False
    """

    def __init__(self, rank=0, log_dir='./logs'):

        # Root logger on DEBUG level
        self.disabled = False
        self.root_logger = logging.getLogger()
        self.root_logger.setLevel(logging.WARNING)
        if not os.path.exists(log_dir):
            os.makedirs(log_dir, exist_ok=True)

        log_name = log_dir + f'/worker-{rank:03d}.log'
        # Console handler on INFO level
        # console_handler = logging.StreamHandler()
        # console_handler.setLevel(logging.INFO)
        log_format = logging.Formatter(
            "%(asctime)s %(name)-25s %(levelname)-9s %(message)s")
        # console_handler.setFormatter(log_format)
        # self.root_logger.addHandler(console_handler)

        self.file_handler = logging.FileHandler(log_name, mode='w')
        self.file_handler.setLevel(logging.WARNING)
        self.file_handler.setFormatter(log_format)
        self.root_logger.addHandler(self.file_handler)
        logging.info("Initialized")
        # if debug == True:
        #     logging.debug("Logger in debug mode")

    def disable(self):
        """Disables all logging."""

        logging.info("Disable logging")
        # logging.disable(level=logging.NOTSET)
        # self.root_logger.setLevel(logging.NOTSET)
        # self.file_handler.setLevel(logging.NOTSET)
        self.root_logger.disabled = True
        self.disabled = True

    def debug(self, string):
        """Set logging to debug verbosity

        Args:
            string (_type_): _description_
        """
        if not self.disabled:
            logging.debug(string)

    def info(self, string):
        """Set logging to info verbosity

        Args:
            string (_type_): _description_
        """
        if not self.disabled:
            logging.info(string)

    def critical(self, string):
        """Set logging to critical verbosity

        Args:
            string (_type_): _description_
        """
        if not self.disabled:
            logging.critical(string)


if WORKER is None:
    WORKER = Worker()




# Copyright 2015 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
**Method to apply a notch filter to a specified impedance source**

:Authors: **Danilo Quartullo**
'''

from __future__ import division

import numpy as np


# FIRST METHOD
def impedance_notches(f_rev, frequencies, imp_source, list_harmonics, list_width_depth):

    halfmajorWidth = list_width_depth[0]
    halfminorWidth = list_width_depth[1]
    majorDepth = list_width_depth[2]
    minorDepth = list_width_depth[3]
    array_left = np.array([])
    indices_final = np.array([])
    frequencies_remainder_array = np.array([])
    Z_remainder_array = np.array([])

    for i in list_harmonics:
        exec("y_" + str(i) + " = np.interp(i*f_rev, frequencies, imp_source)")
        array_left = np.append(array_left, np.array([i * f_rev - halfmajorWidth, i * f_rev + halfmajorWidth]))
        exec("indices_" + str(i) + " = np.where((frequencies >= (i*f_rev-halfmajorWidth))&(frequencies <= (i*f_rev+halfmajorWidth)))[0]")
        indices_final = np.append(indices_final, locals()['indices_' + str(i)])
        frequencies_remainder_array = np.append(frequencies_remainder_array, np.array([i * f_rev - halfminorWidth, i * f_rev, i * f_rev + halfminorWidth]))
        Z_remainder_array = np.append(Z_remainder_array, np.array([locals()['y_' + str(i)] / minorDepth, locals()['y_' + str(i)] / majorDepth, locals()['y_' + str(i)] / minorDepth]))

    left_y = np.interp(array_left, frequencies, imp_source)
    frequencies_remainder = np.delete(frequencies, indices_final)
    Z_remainder = np.delete(imp_source, indices_final)

    frequencies_remainder_array = np.append(frequencies_remainder_array, array_left)
    Z_remainder_array = np.append(Z_remainder_array, left_y)
    frequencies_remainder = np.append(frequencies_remainder, frequencies_remainder_array)
    Z_remainder = np.append(Z_remainder, Z_remainder_array)

    ordered_freq = np.argsort(frequencies_remainder)
    frequencies_remainder = np.take(frequencies_remainder, ordered_freq)
    Z_closed = np.take(Z_remainder, ordered_freq)

    return [frequencies_remainder, Z_closed]



'''
BLonD physics functions, python-only implementations

@author: alasheen, kiliakis
'''

import numpy as np

RNG = np.random.default_rng()


# --------------- Similar to kick.cpp -----------------
def kick(dt: np.ndarray, dE: np.ndarray, voltage: np.ndarray,
         omega_rf: np.ndarray, phi_rf: np.ndarray,
         charge: float, n_rf: int, acceleration_kick: float) -> None:
    '''
    Function to apply RF kick on the particles with sin function
    '''

    voltage_kick = charge * voltage

    for j in range(n_rf):
        dE += voltage_kick[j] * np.sin(omega_rf[j] * dt + phi_rf[j])

    dE[:] += acceleration_kick


def rf_volt_comp(voltages: np.ndarray, omega_rf: np.ndarray, phi_rf: np.ndarray,
                 bin_centers: np.ndarray) -> np.ndarray:
    """Compute rf voltage at each bin.

    Args:
        voltages (np.ndarray): _description_
        omega_rf (np.ndarray): _description_
        phi_rf (np.ndarray): _description_
        bin_centers (np.ndarray): _description_

    Returns:
        np.ndarray: _description_
    """
    rf_voltage = np.zeros(len(bin_centers))

    for j in range(len(voltages)):
        rf_voltage += voltages[j] * np.sin(
            omega_rf[j] * bin_centers + phi_rf[j])

    return rf_voltage
# ---------------------------------------------------


# --------------- Similar to drift.cpp -----------------
def drift(dt: np.ndarray, dE: np.ndarray, solver: str, t_rev: float,
          length_ratio: float, alpha_order, eta_0: float,
          eta_1: float, eta_2: float, alpha_0: float,
          alpha_1: float, alpha_2: float, beta: float, energy: float) -> None:
    '''
    Function to apply drift equation of motion
    '''

    # solver_decoded = solver.decode(encoding='utf_8')

    T = t_rev * length_ratio

    if solver == 'simple':
        coeff = eta_0 / (beta * beta * energy)
        dt += T * coeff * dE

    elif solver == 'legacy':
        coeff = 1. / (beta * beta * energy)
        eta0 = eta_0 * coeff
        eta1 = eta_1 * coeff * coeff
        eta2 = eta_2 * coeff * coeff * coeff

        if alpha_order == 0:
            dt += T * (1. / (1. - eta0 * dE) - 1.)
        elif alpha_order == 1:
            dt += T * (1. / (1. - eta0 * dE
                                - eta1 * dE * dE) - 1.)
        else:
            dt += T * (1. / (1. - eta0 * dE
                                - eta1 * dE * dE
                                - eta2 * dE * dE * dE) - 1.)

    else:
        invbetasq = 1 / (beta * beta)
        invenesq = 1 / (energy * energy)
        # double beam_delta;

        beam_delta = np.sqrt(1. + invbetasq *
                             (dE * dE * invenesq + 2. * dE / energy)) - 1.

        dt += T * (
            (1. + alpha_0 * beam_delta +
             alpha_1 * (beam_delta * beam_delta) +
             alpha_2 * (beam_delta * beam_delta * beam_delta)) *
            (1. + dE / energy) / (1. + beam_delta) - 1.)
# ---------------------------------------------------


# --------------- Similar to histogram.cpp -----------------
def slice_beam(dt: np.ndarray, profile: np.ndarray,
          cut_left: float, cut_right: float) -> None:
    """Slice the time coordinate of the beam.

    Args:
        dt (np.ndarray): _description_
        profile (np.ndarray): _description_
        cut_left (float): _description_
        cut_right (float): _description_
    """
    profile[:] = np.histogram(dt, bins=len(profile),
                              range=(cut_left, cut_right))[0]


def slice_smooth(dt: np.ndarray, profile: np.ndarray,
                 cut_left: float, cut_right: float) -> None:
    """Smooth slice method.

    Args:
        dt (np.ndarray): _description_
        profile (np.ndarray): _description_
        cut_left (float): _description_
        cut_right (float): _description_
    """
    # Constants init
    n_slices = len(profile)
    bin_width = (cut_right - cut_left) / n_slices
    inv_bin_width = 1. / bin_width
    const1 = (cut_left + bin_width * 0.5)
    const2 = (cut_right - bin_width * 0.5)

    profile[:] = 0.0

    for a in dt:
        if (a < const1) or (a > const2):
            continue

        fbin = (a - cut_left) * inv_bin_width
        ffbin = int(fbin)
        distToCenter = fbin - ffbin
        if distToCenter > 0.5:
            fffbin = int(fbin + 1)
        else:
            fffbin = int(fbin - 1)
        profile[ffbin] += (0.5 - distToCenter)
        profile[fffbin] += (0.5 + distToCenter)
# ---------------------------------------------------


# --------------- Similar to linear_interp_kick.cpp -----------------
def linear_interp_kick(dt: np.ndarray, dE: np.ndarray, voltage: np.ndarray,
                       bin_centers: np.ndarray, charge: float,
                       acceleration_kick: float) -> None:
    """Interpolated kick method.

    Args:
        dt (np.ndarray): _description_
        dE (np.ndarray): _description_
        voltage (np.ndarray): _description_
        bin_centers (np.ndarray): _description_
        charge (float): _description_
        acceleration_kick (float): _description_
    """
    n_slices = len(bin_centers)
    inv_bin_width = (n_slices - 1) / (bin_centers[-1] - bin_centers[0])

    fbin = np.floor((dt - bin_centers[0]) * inv_bin_width).astype(np.int32)

    helper1 = charge * (voltage[1:] - voltage[:-1]) * inv_bin_width
    helper2 = (charge * voltage[:-1] -
               bin_centers[:-1] * helper1) + acceleration_kick

    for i in range(len(dt)):
        # fbin = int(np.floor((dt[i]-bin_centers[0])*inv_bin_width))
        if (fbin[i] >= 0) and (fbin[i] < n_slices - 1):
            dE[i] += dt[i] * helper1[fbin[i]] + helper2[fbin[i]]
# ---------------------------------------------------


# --------------- Similar to synchrotron_radiation.cpp -----------------
def synchrotron_radiation(dE: np.ndarray, U0: float,
                          n_kicks: int, tau_z: float) -> None:
    """Apply SR

    Args:
        dE (np.ndarray): _description_
        U0 (float): _description_
        n_kicks (int): _description_
        tau_z (float): _description_
    """
    for _ in range(n_kicks):
        dE += -(2.0 / tau_z / n_kicks * dE + U0 / n_kicks)


def synchrotron_radiation_full(dE: np.ndarray, U0: float,
                               n_kicks: int, tau_z: float,
                               sigma_dE: float, energy: float) -> None:
    """Apply SR with quantum excitation

    Args:
        dE (np.ndarray): _description_
        U0 (float): _description_
        n_kicks (int): _description_
        tau_z (float): _description_
        sigma_dE (float): _description_
        energy (float): _description_
    """

    for _ in range(n_kicks):
        dE += -(2.0 / tau_z / n_kicks * dE +
                U0 / n_kicks - 2.0 * sigma_dE /
                np.sqrt(tau_z * n_kicks) * energy *
                RNG.standard_normal(len(dE)))


def set_random_seed(seed: int) -> None:
    """Set the seed of the RNG used in synchrotron radiation

    Args:
        seed (int): _description_
    """
    global RNG
    # Re-initialize the RNG with new seed
    RNG = np.random.default_rng(seed)
# ---------------------------------------------------


# --------------- Similar to music_track.cpp -----------------
def music_track(dt: np.ndarray, dE: np.ndarray, induced_voltage: np.ndarray,
                array_parameters: np.ndarray, alpha: float, omega_bar: float,
                const: float, coeff1: float, coeff2: float,
                coeff3: float, coeff4: float) -> None:
    '''
    This function calculates the single-turn induced voltage and updates the
    energies of the particles.

    Parameters
    ----------
    beam_dt : float array
        Longitudinal coordinates [s]
    beam_dE : float array
        Initial energies [V]
    induced_voltage : float array
        array used to store the output of the computation
    array_parameters : float array
        See documentation in music.py
    n_macroparticles : int
        number of macro-particles
    alpha, omega_bar, cnst, coeff1, coeff2, coeff3, coeff4 : floats
        See documentation in music.py

    Returns
    -------
    induced_voltage : float array
        Computed induced voltage.
    beam_dE : float array
        Array of energies updated.
    '''

    indices_sorted = np.argsort(dt)
    dt = dt[indices_sorted]
    dE = dE[indices_sorted]

    # MuSiC algorithm
    dE[0] += induced_voltage[0]
    input_first_component = 1.0
    input_second_component = 0.0

    for i in range(len(dt) - 1):

        time_difference = dt[i + 1] - dt[i]

        exp_term = np.exp(-alpha * time_difference)
        cos_term = np.cos(omega_bar * time_difference)
        sin_term = np.sin(omega_bar * time_difference)

        product_first_component = exp_term * \
            ((cos_term + coeff1 * sin_term) * input_first_component
                + coeff2 * sin_term * input_second_component)
        product_second_component = exp_term * \
            (coeff3 * sin_term * input_first_component
                + (cos_term + coeff4 * sin_term) * input_second_component)

        induced_voltage[i + 1] = const * (0.5 + product_first_component)
        dE[i + 1] += induced_voltage[i + 1]

        input_first_component = product_first_component + 1.0
        input_second_component = product_second_component

    array_parameters[0] = input_first_component
    array_parameters[1] = input_second_component
    array_parameters[3] = dt[-1]


def music_track_multiturn(dt: np.ndarray, dE: np.ndarray, induced_voltage: np.ndarray,
                          array_parameters: np.ndarray, alpha: float, omega_bar: float,
                          const: float, coeff1: float, coeff2: float,
                          coeff3: float, coeff4: float) -> None:
    """This function calculates the multi-turn induced voltage and updates the
    energies of the particles.
    Parameters and Returns as for music_track.

    Args:
        dt (np.ndarray): _description_
        dE (np.ndarray): _description_
        induced_voltage (np.ndarray): _description_
        array_parameters (np.ndarray): _description_
        alpha (float): _description_
        omega_bar (float): _description_
        const (float): _description_
        coeff1 (float): _description_
        coeff2 (float): _description_
        coeff3 (float): _description_
        coeff4 (float): _description_
    """
    indices_sorted = np.argsort(dt)
    dt = dt[indices_sorted]
    dE = dE[indices_sorted]
    time_difference_0 = dt[0] + array_parameters[2] - array_parameters[3]
    exp_term = np.exp(-alpha * time_difference_0)
    cos_term = np.cos(omega_bar * time_difference_0)
    sin_term = np.sin(omega_bar * time_difference_0)

    product_first_component = exp_term * (
        (cos_term + coeff1 * sin_term)
        * array_parameters[0]
        + coeff2 * sin_term * array_parameters[1])

    product_second_component = exp_term * (
        coeff3 * sin_term * array_parameters[0]
        + (cos_term + coeff4 * sin_term) * array_parameters[1])

    induced_voltage[0] = const * (0.5 + product_first_component)

    dE[0] += induced_voltage[0]
    input_first_component = product_first_component + 1.0
    input_second_component = product_second_component

    # MuSiC algorithm for the current turn

    for i in range(len(dt) - 1):

        time_difference = dt[i + 1] - dt[i]

        exp_term = np.exp(-alpha * time_difference)
        cos_term = np.cos(omega_bar * time_difference)
        sin_term = np.sin(omega_bar * time_difference)

        product_first_component = exp_term * \
            ((cos_term + coeff1 * sin_term) * input_first_component
                + coeff2 * sin_term * input_second_component)
        product_second_component = exp_term * \
            (coeff3 * sin_term * input_first_component
                + (cos_term + coeff4 * sin_term) * input_second_component)

        induced_voltage[i + 1] = const * (0.5 + product_first_component)
        dE[i + 1] += induced_voltage[i + 1]

        input_first_component = product_first_component + 1.0
        input_second_component = product_second_component

    array_parameters[0] = input_first_component
    array_parameters[1] = input_second_component
    array_parameters[3] = dt[-1]
# ---------------------------------------------------


# --------------- Similar to fast_resonator.cpp -----------------
def fast_resonator(R_S: np.ndarray, Q: np.ndarray, frequency_array: np.ndarray,
                   frequency_R: np.ndarray, impedance: np.ndarray = None) -> np.ndarray:
    '''
    This function takes as an input a list of resonators parameters and
    computes the impedance in an optimised way.

    Parameters
    ----------
    frequencies: float array
        array of frequency in Hz
    shunt_impedances: float array
        array of shunt impedances in Ohm
    Q_values: float array
        array of quality factors
    resonant_frequencies: float array
        array of resonant frequency in Hz
    n_resonators: int
        number of resonantors
    n_frequencies: int
        length of the array 'frequencies'

    Returns
    -------
    impedanceReal: float array
        real part of the impedance
    impedanceImag: float array
        imaginary part of the impedance
    '''
    if impedance is None:
        impedance = np.zeros(len(frequency_array), dtype=complex)

    for i in range(len(R_S)):
        impedance[1:] += R_S[i] / (1 + 1j * Q[i] * (frequency_array[1:] / frequency_R[i] -
                                                    frequency_R[i] / frequency_array[1:]))

    return impedance
# ---------------------------------------------------


# --------------- Similar to beam_phase.cpp -----------------
def beam_phase(bin_centers: np.ndarray, profile: np.ndarray,
               alpha: float, omegarf: float,
               phirf: float, bin_size: float) -> float:
    scoeff = np.trapz(np.exp(alpha * (bin_centers))
                      * np.sin(omegarf * bin_centers + phirf)
                      * profile, dx=bin_size)
    ccoeff = np.trapz(np.exp(alpha * (bin_centers))
                      * np.cos(omegarf * bin_centers + phirf)
                      * profile, dx=bin_size)

    return scoeff / ccoeff


def beam_phase_fast(bin_centers: np.ndarray, profile: np.ndarray,
                    omegarf: float, phirf: float, bin_size: float) -> float:
    scoeff = np.trapz(profile * np.sin(omegarf * bin_centers + phirf),
                      dx=bin_size)
    ccoeff = np.trapz(profile * np.cos(omegarf * bin_centers + phirf),
                      dx=bin_size)

    return scoeff / ccoeff
# ---------------------------------------------------


# --------------- Similar to sparse_histogram.cpp -----------------
def sparse_histogram(dt: np.ndarray, profile: np.ndarray,
                     cut_left: np.ndarray, cut_right: np.ndarray,
                     bunch_indexes: np.ndarray, n_slices_bucket: int) -> None:
    '''
    Optimised routine that calculates the histogram for a sparse beam
    Author: Juan F. Esteban Mueller, Danilo Quartullo, Alexandre Lasheen, Markus Schwarz
    '''
    # Only valid for cut_edges = edges
    inv_bucket_length = 1.0 / (cut_right[0] - cut_left[0])
    inv_bin_width = inv_bucket_length * n_slices_bucket

    profile[:] = 0.0

    for a in dt:
        if ((a < cut_left[0]) or (a > cut_right[-1])):
            continue
        # Find bucket in which the particle is and its index
        fbunch = int((a - cut_left[0]) * inv_bucket_length)
        i_bucket = int(bunch_indexes[fbunch])
        if i_bucket == -1:
            continue
        # Find the bin inside the corresponding bucket
        fbin = int((a - cut_left[i_bucket]) * inv_bin_width)
        profile[i_bucket, fbin] += 1.0
# ---------------------------------------------------


# --------------- Similar to tomoscope.cpp -----------------
def distribution_from_tomoscope(dt: np.ndarray, dE: np.ndarray, probDistr: np.ndarray,
                                seed: int, profLen: int,
                                cutoff: float, x0: float, y0: float,
                                dtBin: float, dEBin: float) -> None:
    '''
    Generation of particle distribution from probability density
    Author: Helga Timko
    '''
    import bisect

    # Initialize random seed
    np.random.seed(seed)

    # Initialize variables
    cutoff2 = cutoff * cutoff
    dtMin = -1.0 * x0 * dtBin
    dEMin = -1.0 * y0 * dEBin

    # Calculate cumulative probability
    # profLen2 = np.uint32(profLen * profLen)
    cumulDistr = np.cumsum(probDistr)

    # Normalize probability distribution
    totProb = cumulDistr[-1]
    cumulDistr = cumulDistr / totProb

    # Generate particle coordinates

    n = 0
    while n < len(dt):
        randProb = np.random.random()
        m = bisect.bisect(cumulDistr, randProb)
        # for m in range(profLen2):
        #     if randProb < cumulDistr[m]:
        #         break
        i = int(m / profLen)
        k = m % profLen

        iPos = i + np.random.random() - 0.5
        kPos = k + np.random.random() - 0.5

        if ((iPos - x0) * (iPos - x0) + (kPos - y0) * (kPos - y0)) < cutoff2:
            dt[n] = dtMin + iPos * dtBin
            dE[n] = dEMin + kPos * dEBin
            n += 1
# ---------------------------------------------------



'''
@author: Konstantinos Iliakis, George Tsapatsaris
'''

import cupy as cp
import numpy as np

from ..utils import precision
from . import GPU_DEV


def rf_volt_comp(voltage, omega_rf, phi_rf, bin_centers):
    """Calculate the rf voltage at each profile bin

    Args:
        voltage (float array): _description_
        omega_rf (float array): _description_
        phi_rf (float array): _description_
        bin_centers (float array): _description_

    Returns:
        float array: the calculated rf_voltage
    """

    rf_volt_comp_kernel = GPU_DEV.mod.get_function("rf_volt_comp")

    
    assert voltage.dtype == precision.real_t
    assert omega_rf.dtype == precision.real_t
    assert phi_rf.dtype == precision.real_t
    assert bin_centers.dtype == precision.real_t

    rf_voltage = cp.zeros(bin_centers.size, precision.real_t)

    rf_volt_comp_kernel(args=(voltage, omega_rf, phi_rf, bin_centers,
                              np.int32(voltage.size), np.int32(bin_centers.size), rf_voltage),
                        block=GPU_DEV.block_size, grid=GPU_DEV.grid_size)
    return rf_voltage


def kick(dt, dE, voltage, omega_rf, phi_rf, charge, n_rf, acceleration_kick):
    """Apply the energy kick

    Args:
        dt (float array): the time coordinate
        dE (float array): the energy coordinate
        voltage (float array): _description_
        omega_rf (float array): _description_
        phi_rf (float array): _description_
        charge (float): _description_
        n_rf (int): _description_
        acceleration_kick (float): _description_
    """
    kick_kernel = GPU_DEV.mod.get_function("simple_kick")

    assert dt.dtype == precision.real_t
    assert dE.dtype == precision.real_t
    assert omega_rf.dtype == precision.real_t
    assert phi_rf.dtype == precision.real_t

    voltage_kick = cp.empty(voltage.size, precision.real_t)
    voltage_kick = charge * voltage
    """
    voltage_kick, omega_rf and phi_rf are not correctly indexed inside
    inside the kernel.  As a result, the wrong kick is applied when
    using multiple RF harmonics.  Ideally, the arrays should be correctly
    indexed within the kernel, but that will require a fairly heavy
    change to a lot of the code.  As a short term solution, casting the
    three with cp.array will work, and adds about 30 us per call.  With
    1E6 particles and two harmonics, that's about a 10% increase in the
    runtime of this function.
    """
    #TODO: Correct indexing of voltage, omega and phi
    kick_kernel(args=(dt,
                      dE,
                      np.int32(n_rf),
                      cp.array(voltage_kick),
                      cp.array(omega_rf),
                      cp.array(phi_rf),
                      np.int32(dt.size),
                      precision.real_t(acceleration_kick)),
                block=GPU_DEV.block_size, grid=GPU_DEV.grid_size)


def drift(dt, dE, solver, t_rev, length_ratio, alpha_order, eta_0,
          eta_1, eta_2, alpha_0, alpha_1, alpha_2, beta, energy):
    """Apply the time drift function.

    Args:
        dt (_type_): _description_
        dE (_type_): _description_
        solver (_type_): _description_
        t_rev (_type_): _description_
        length_ratio (_type_): _description_
        alpha_order (_type_): _description_
        eta_0 (_type_): _description_
        eta_1 (_type_): _description_
        eta_2 (_type_): _description_
        alpha_0 (_type_): _description_
        alpha_1 (_type_): _description_
        alpha_2 (_type_): _description_
        beta (_type_): _description_
        energy (_type_): _description_
    """
    drift_kernel = GPU_DEV.mod.get_function("drift")

    solver_to_int = {
        'simple': 0,
        'legacy': 1,
        'exact': 2,
    }
    solver = solver_to_int[solver]

    if not isinstance(t_rev, precision.real_t):
        t_rev = precision.real_t(t_rev)

    drift_kernel(args=(dt, dE, solver,
                       precision.real_t(t_rev), precision.real_t(length_ratio),
                       precision.real_t(alpha_order), precision.real_t(eta_0),
                       precision.real_t(eta_1), precision.real_t(eta_2),
                       precision.real_t(alpha_0), precision.real_t(alpha_1),
                       precision.real_t(alpha_2),
                       precision.real_t(beta), precision.real_t(energy),
                       np.int32(dt.size)),
                 block=GPU_DEV.block_size, grid=GPU_DEV.grid_size)


def linear_interp_kick(dt, dE, voltage,
                       bin_centers, charge,
                       acceleration_kick):
    """An accelerated version of the kick function.

    Args:
        dt (_type_): _description_
        dE (_type_): _description_
        voltage (_type_): _description_
        bin_centers (_type_): _description_
        charge (_type_): _description_
        acceleration_kick (_type_): _description_
    """
    gm_linear_interp_kick_help = GPU_DEV.mod.get_function("lik_only_gm_copy")
    gm_linear_interp_kick_comp = GPU_DEV.mod.get_function("lik_only_gm_comp")

    assert dt.dtype == precision.real_t
    assert dE.dtype == precision.real_t
    assert voltage.dtype == precision.real_t
    assert bin_centers.dtype == precision.real_t

    macros = dt.size
    slices = bin_centers.size

    glob_vkick_factor = cp.empty(2 * (slices - 1), precision.real_t)
    gm_linear_interp_kick_help(args=(dt,
                                     dE,
                                     voltage,
                                     bin_centers,
                                     precision.real_t(charge),
                                     np.int32(slices),
                                     np.int32(macros),
                                     precision.real_t(acceleration_kick),
                                     glob_vkick_factor),
                               grid=GPU_DEV.grid_size, block=GPU_DEV.block_size)

    gm_linear_interp_kick_comp(args=(dt,
                                     dE,
                                     voltage,
                                     bin_centers,
                                     precision.real_t(charge),
                                     np.int32(slices),
                                     np.int32(macros),
                                     precision.real_t(acceleration_kick),
                                     glob_vkick_factor),
                               grid=GPU_DEV.grid_size, block=GPU_DEV.block_size)



def slice_beam(dt, profile, cut_left, cut_right):
    """Constant space slicing with a constant frame.

    Args:
        dt (_type_): _description_
        profile (_type_): _description_
        cut_left (_type_): _description_
        cut_right (_type_): _description_
    """
    sm_histogram = GPU_DEV.mod.get_function("sm_histogram")
    hybrid_histogram = GPU_DEV.mod.get_function("hybrid_histogram")



    assert dt.dtype == precision.real_t

    n_slices = profile.size
    profile.fill(0)

    if not isinstance(cut_left, float):
        cut_left = float(cut_left)
    if not isinstance(cut_right, float):
        cut_right = float(cut_right)

    if 4 * n_slices < GPU_DEV.attributes['MaxSharedMemoryPerBlock']:
        sm_histogram(args=(dt, profile, precision.real_t(cut_left),
                           precision.real_t(cut_right), np.uint32(n_slices),
                           np.uint32(dt.size)),
                     grid=GPU_DEV.grid_size, block=GPU_DEV.block_size, shared_mem=4 * n_slices)
    else:
        hybrid_histogram(args=(dt, profile, precision.real_t(cut_left),
                               precision.real_t(cut_right), np.uint32(n_slices),
                               np.uint32(dt.size), np.int32(
            GPU_DEV.attributes['MaxSharedMemoryPerBlock'] / 4)),
            grid=GPU_DEV.grid_size, block=GPU_DEV.block_size,
            shared_mem=GPU_DEV.attributes['MaxSharedMemoryPerBlock'])


def synchrotron_radiation(dE, U0, n_kicks, tau_z):
    """Track particles with SR only (without quantum excitation)

    Args:
        dE (_type_): _description_
        U0 (_type_): _description_
        n_kicks (_type_): _description_
        tau_z (_type_): _description_
    """
    synch_rad = GPU_DEV.mod.get_function("synchrotron_radiation")

    assert dE.dtype == precision.real_t

    synch_rad(args=(dE, precision.real_t(U0 / n_kicks), np.int32(dE.size),
                    precision.real_t(tau_z * n_kicks),
                    np.int32(n_kicks)),
              block=GPU_DEV.block_size, grid=GPU_DEV.grid_size)


def synchrotron_radiation_full(dE, U0, n_kicks, tau_z, sigma_dE, energy):
    """Track particles with SR and quantum excitation.

    Args:
        dE (_type_): _description_
        U0 (_type_): _description_
        n_kicks (_type_): _description_
        tau_z (_type_): _description_
        sigma_dE (_type_): _description_
        energy (_type_): _description_
    """
    synch_rad_full = GPU_DEV.mod.get_function("synchrotron_radiation_full")

    assert dE.dtype == precision.real_t

    synch_rad_full(args=(dE, precision.real_t(U0 / n_kicks), np.int32(dE.size),
                         precision.real_t(sigma_dE),
                         precision.real_t(tau_z * n_kicks),
                         precision.real_t(energy), np.int32(n_kicks)),
                   block=GPU_DEV.block_size, grid=GPU_DEV.grid_size)


@cp.fuse(kernel_name='beam_phase_helper')
def __beam_phase_helper(bin_centers, profile, alpha, omega_rf, phi_rf):
    """Helper function, used by beam_phase

    Args:
        bin_centers (_type_): _description_
        profile (_type_): _description_
        alpha (_type_): _description_
        omega_rf (_type_): _description_
        phi_rf (_type_): _description_

    Returns:
        _type_: _description_
    """
    base = cp.exp(alpha * bin_centers) * profile
    a = omega_rf * bin_centers + phi_rf
    return base * cp.sin(a), base * cp.cos(a)


def beam_phase(bin_centers, profile, alpha, omega_rf, phi_rf, bin_size):
    """Beam phase measured at the main RF frequency and phase. The beam is
       convolved with the window function of the band-pass filter of the
       machine. The coefficients of sine and cosine components determine the
       beam phase, projected to the range -Pi/2 to 3/2 Pi. Note that this beam
       phase is already w.r.t. the instantaneous RF phase.*

    Args:
        bin_centers (_type_): _description_
        profile (_type_): _description_
        alpha (_type_): _description_
        omega_rf (_type_): _description_
        phi_rf (_type_): _description_
        bin_size (_type_): _description_

    Returns:
        _type_: _description_
    """
    assert bin_centers.dtype == precision.real_t
    assert profile.dtype == precision.real_t

    array1, array2 = __beam_phase_helper(
        bin_centers, profile, alpha, omega_rf, phi_rf)
    # due to the division, the bin_size is not needed
    scoeff = cp.trapz(array1, dx=1)
    ccoeff = cp.trapz(array2, dx=1)

    return float(scoeff / ccoeff)


@cp.fuse(kernel_name='beam_phase_fast_helper')
def __beam_phase_fast_helper(bin_centers, profile, omega_rf, phi_rf):
    """Helper function used by beam_phase_fast

    Args:
        bin_centers (_type_): _description_
        profile (_type_): _description_
        omega_rf (_type_): _description_
        phi_rf (_type_): _description_

    Returns:
        _type_: _description_
    """
    arr = omega_rf * bin_centers + phi_rf
    return profile * cp.sin(arr), profile * cp.cos(arr)


def beam_phase_fast(bin_centers, profile, omega_rf, phi_rf, bin_size):
    """Simplified, faster variation of the beam_phase function

    Args:
        bin_centers (_type_): _description_
        profile (_type_): _description_
        omega_rf (_type_): _description_
        phi_rf (_type_): _description_
        bin_size (_type_): _description_

    Returns:
        _type_: _description_
    """
    assert bin_centers.dtype == precision.real_t
    assert profile.dtype == precision.real_t

    array1, array2 = __beam_phase_fast_helper(
        bin_centers, profile, omega_rf, phi_rf)
    # due to the division, the bin_size is not needed
    scoeff = cp.trapz(array1, dx=1)
    ccoeff = cp.trapz(array2, dx=1)

    return float(scoeff / ccoeff)



# Copyright 2016 CERN. This software is distributed under the
# terms of the GNU General Public License version 3 (GPL Version 3),
# copied verbatim in the file LICENSE.md.
# In applying this license, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
** CERN Tomoscope-related functions to generate particle distribution or 
probability density.**

:Authors: **Helga Timko**
'''


from __future__ import division, print_function

import linecache
import os
from builtins import str

import h5py as hp
import matplotlib.cm as cm
import numpy as np
from matplotlib import pyplot as plt

from ..utils import bmath as bm


def distribution_from_tomoscope_data(dataDir, nPart, cutoff=1000, seed=1234,
                                     plotFig=True, saveDistr=False):
    '''
    'dataDir' is the directory of *directories* of offline-processed tomoscope 
    data containing 'plotinfo.data' and 'image001.data'
    Use 'cutoff' to eliminate measurement noise (use values around 50-100).
    Use 'seed' to change the random number seed.
    Use 'plotFig = True' to plot and save figures.
    Use 'saveDistr = True' to save particle coordinates.    
    '''

    # Directory in which plots will be stored
    distrDir = dataDir + '\\Distributions_' + \
        str(nPart) + 'particles_cutoff' + str(cutoff)
    # Try to create directory
    try:
        os.makedirs(distrDir)
    # Check whether already exists/creation failed
    except OSError:
        if os.path.exists(distrDir):
            pass
        else:
            raise

    # Loop over directories to be analysed
    for directory in os.listdir(dataDir):

        if 'Distributions' not in directory:

            print("Analysing data of directory %s" % directory)

            # Read tomoscope settings
            plotInfo = dataDir + '\\' + directory + '\\plotinfo.data'
            profLen = np.uint(linecache.getline(plotInfo, 4)[17:-1])
            dtBin = np.double(linecache.getline(plotInfo, 6)[9:-1])
            dEBin = np.double(linecache.getline(plotInfo, 8)[9:-1])
            x0 = np.double(linecache.getline(plotInfo, 12)[8:-1])
            y0 = np.double(linecache.getline(plotInfo, 13)[8:-1])

            # Read probability density from file
            probDistr = np.loadtxt(dataDir + '\\' + directory + '\\image001.data',
                                   dtype=np.double, unpack=True)
            probDistr = np.ascontiguousarray(probDistr)

            # Reconstruct particle distribution from probability distribution
            dt = np.empty(nPart)
            dE = np.empty(nPart)

            bm.distribution_from_tomoscope(dt, dE, probDistr, seed, profLen,
                                           cutoff, x0, y0, dtBin, dEBin)

            if plotFig:

                # Settings for plots
                plt.rc('axes', labelsize=14, labelweight='normal')
                plt.rc('lines', linewidth=1.5, markersize=6)
                plt.rc('font', family='sans-serif')
                plt.rc('legend', fontsize=12)

                # Plot distribution
                plt.hist2d(dt, dE, (profLen, profLen), cmap=cm.jet)
                plt.colorbar()
                plt.xlabel('Time offset [s]')
                plt.ylabel('Energy offset [eV]')
                plt.savefig(distrDir + '\\' + directory + '.png')
                plt.clf()

            if saveDistr:

                h5File = hp.File(distrDir + '\\' + directory + '.h5', 'w')

                # Create group
                h5File.require_group('Beam')
                h5Group = h5File['Beam']

                # Create & write datasets
                h5Group.create_dataset("dt", shape=(nPart,), dtype='f',
                                       compression="gzip", compression_opts=9)
                h5Group["dt"][:] = dt
                h5Group.create_dataset("dE", shape=(nPart,), dtype='f',
                                       compression="gzip", compression_opts=9)
                h5Group["dE"][:] = dE
                h5File.close()



'''
Basic methods and objects related to the computational core.

@author: Konstantinos Iliakis
@date: 25.05.2023
'''

import ctypes as ct
import numpy as np


class PrecisionClass:
    """Singleton class. Holds information about the floating point precision of the calculations.
    """
    __instance = None

    def __init__(self, _precision='double'):
        """Constructor

        Args:
            _precision (str, optional): _description_. Defaults to 'double'.
        """
        if PrecisionClass.__instance is not None:
            return

        PrecisionClass.__instance = self
        self.set(_precision)

    def set(self, _precision='double'):
        """Set the precision to single or double.

        Args:
            _precision (str, optional): _description_. Defaults to 'double'.
        """
        if _precision in ['single', 's', '32', 'float32', 'float', 'f']:
            self.str = 'single'
            self.real_t = np.float32
            self.c_real_t = ct.c_float
            self.complex_t = np.complex64
            self.num = 1
        elif _precision in ['double', 'd', '64', 'float64']:
            self.str = 'double'
            self.real_t = np.float64
            self.c_real_t = ct.c_double
            self.complex_t = np.complex128
            self.num = 2


class c_complex128(ct.Structure):
    """128-bit (64+64) Complex number, compatible with std::complex layout

    Args:
        ct (_type_): _description_

    Returns:
        _type_: _description_
    """
    _fields_ = [("real", ct.c_double), ("imag", ct.c_double)]

    def __init__(self, pycomplex):
        """Init from Python complex

        Args:
            pycomplex (_type_): _description_
        """
        self.real = pycomplex.real.astype(np.float64, order='C')
        self.imag = pycomplex.imag.astype(np.float64, order='C')

    def to_complex(self):
        """Convert to Python complex

        Returns:
            _type_: _description_
        """
        return self.real + (1.j) * self.imag


class c_complex64(ct.Structure):
    """64-bit (32+32) Complex number, compatible with std::complex layout

    Args:
        ct (_type_): _description_

    Returns:
        _type_: _description_
    """
    _fields_ = [("real", ct.c_float), ("imag", ct.c_float)]

    def __init__(self, pycomplex):
        """Init from Python complex

        Args:
            pycomplex (_type_): _description_
        """
        self.real = pycomplex.real.astype(np.float32, order='C')
        self.imag = pycomplex.imag.astype(np.float32, order='C')

    def to_complex(self):
        """Convert to Python complex

        Returns:
            _type_: _description_
        """
        return self.real + (1.j) * self.imag


def c_real(scalar):
    """Convert input to default precision.

    Args:
        x (_type_): _description_

    Returns:
        _type_: _description_
    """
    if precision.num == 1:
        return ct.c_float(scalar)
    return ct.c_double(scalar)


def c_complex(scalar):
    """Convert input to default precision.

    Args:
        x (_type_): _description_

    Returns:
        _type_: _description_
    """
    if precision.num == 1:
        return c_complex64(scalar)
    return c_complex128(scalar)


# By default use double precision
precision = PrecisionClass('double')



'''
BLonD physics functions, numba implementations
'''

import numpy as np

from numba import jit
from numba import prange
from numba import get_num_threads, get_thread_id
import math
import random 


# --------------- Similar to kick.cpp -----------------
@jit(nopython=True, nogil=True, fastmath=True, parallel=True, cache=True)
def kick(dt: np.ndarray, dE: np.ndarray, voltage: np.ndarray,
         omega_rf: np.ndarray, phi_rf: np.ndarray,
         charge: float, n_rf: int, acceleration_kick: float) -> None:
    '''
    Function to apply RF kick on the particles with sin function
    '''
    voltage_kick = charge * voltage

    for j in range(len(voltage)):
        if j == 0:
            add_kick = acceleration_kick
        else:
            add_kick = 0.0
        for i in prange(len(dt)):
            dE[i] = dE[i] + voltage_kick[j] * \
                np.sin(omega_rf[j] * dt[i] + phi_rf[j]) + add_kick


@jit(nopython=True, nogil=True, fastmath=True, parallel=True, cache=True)
def rf_volt_comp(voltages: np.ndarray, omega_rf: np.ndarray, phi_rf: np.ndarray,
                 bin_centers: np.ndarray) -> np.ndarray:
    """Compute rf voltage at each bin.

    Args:
        voltages (np.ndarray): _description_
        omega_rf (np.ndarray): _description_
        phi_rf (np.ndarray): _description_
        bin_centers (np.ndarray): _description_

    Returns:
        np.ndarray: _description_
    """
    rf_voltage = np.zeros(len(bin_centers))

    for j in range(len(voltages)):
        for i in prange(len(bin_centers)):
            rf_voltage[i] += voltages[j] * \
                np.sin(omega_rf[j] * bin_centers[i] + phi_rf[j])

    return rf_voltage
# ---------------------------------------------------


# --------------- Similar to drift.cpp -----------------
@jit(nopython=True, fastmath=True, parallel=True, cache=True)
def drift(dt: np.ndarray, dE: np.ndarray, solver: str, t_rev: float,
          length_ratio: float, alpha_order, eta_0: float,
          eta_1: float, eta_2: float, alpha_0: float,
          alpha_1: float, alpha_2: float, beta: float, energy: float) -> None:
    '''
    Function to apply drift equation of motion
    0 == 'simple'
    1 == 'legacy'
    2 == 'exact'
    '''

    T = t_rev * length_ratio

    if solver == 'simple':
        coeff = eta_0 / (beta * beta * energy)
        for i in prange(len(dt)):
            dt[i] += T * coeff * dE[i]

    elif solver == 'legacy':
        coeff = 1. / (beta * beta * energy)
        eta0 = eta_0 * coeff
        eta1 = eta_1 * coeff * coeff
        eta2 = eta_2 * coeff * coeff * coeff

        if alpha_order == 0:
            for i in prange(len(dt)):
                dt[i] += T * (1. / (1. - eta0 * dE[i]) - 1.)
        elif alpha_order == 1:
            for i in prange(len(dt)):
                dt[i] += T * (1. / (1. - eta0 * dE[i]
                                    - eta1 * dE[i] * dE[i]) - 1.)
        else:
            for i in prange(len(dt)):
                dt[i] += T * (1. / (1. - eta0 * dE[i]
                                    - eta1 * dE[i] * dE[i]
                                    - eta2 * dE[i] * dE[i] * dE[i]) - 1.)

    else:
        invbetasq = 1 / (beta * beta)
        invenesq = 1 / (energy * energy)
        for i in prange(len(dt)):
            beam_delta = np.sqrt(1. + invbetasq *
                                 (dE[i] * dE[i] * invenesq + 2. * dE[i] / energy)) - 1.

            dt[i] += T * (
                (1. + alpha_0 * beam_delta +
                 alpha_1 * (beam_delta * beam_delta) +
                 alpha_2 * (beam_delta * beam_delta * beam_delta)) *
                (1. + dE[i] / energy) / (1. + beam_delta) - 1.)
# ---------------------------------------------------


# --------------- Similar to histogram.cpp -----------------
@jit(nopython=True, nogil=True, fastmath=True, parallel=True, cache=True)
def slice_beam(dt: np.ndarray, profile: np.ndarray,
               cut_left: float, cut_right: float) -> None:
    """Slice the time coordinate of the beam.

    Args:
        dt (np.ndarray): _description_
        profile (np.ndarray): _description_
        cut_left (float): _description_
        cut_right (float): _description_
    """

    n_slices = len(profile)
    n_parts = len(dt)
    inv_bin_width = n_slices / (cut_right - cut_left)
    n_threads = get_num_threads()

    # Per thread private profile to avoid cross-thread synchronization
    local_profile = np.zeros((n_threads, n_slices),
                             dtype=np.int32)

    # Operate in chunks of 512 particles to avoid calling the expensive
    # get_thread_id() function too often
    STEP = 512
    local_target_bin = np.empty((n_threads, STEP), dtype=np.int32)
    total_steps = math.ceil(n_parts / STEP)

    for i in prange(total_steps):
        thr_id = get_thread_id()
        start_i = i*STEP
        loop_count = min(STEP, n_parts - start_i)
        local_target_bin[thr_id][:loop_count] = np.floor(
            (dt[start_i:start_i+loop_count] - cut_left) * inv_bin_width)

        for j in range(loop_count):
            if local_target_bin[thr_id][j] >= 0 and local_target_bin[thr_id][j] < n_slices:
                local_profile[thr_id, local_target_bin[thr_id][j]] += 1

    # reduce the private profiles to the global profile
    for i in prange(n_slices):
        profile[i] = 0.0
        for j in range(n_threads):
            profile[i] += local_profile[j, i]

    # profile[:] = 0.0

    # profile_len = len(profile)
    # inv_bin_width = profile_len / (cut_right - cut_left)
    # target_bin = np.empty(len(dt), dtype=np.int32)
    # for i in prange(len(dt)):
    #     target_bin[i] = int(np.floor((dt[i] - cut_left) * inv_bin_width))

    # for tbin in target_bin:
    #     if tbin >= 0 and tbin < profile_len:
    #         profile[tbin] += 1.0

    # profile[:] = np.histogram(dt, bins=len(profile),
    #                          range=(cut_left, cut_right))[0]


@jit(nopython=True, fastmath=True, parallel=True, cache=True)
def slice_smooth(dt: np.ndarray, profile: np.ndarray,
                 cut_left: float, cut_right: float) -> None:
    """Smooth slice method.

    Args:
        dt (np.ndarray): _description_
        profile (np.ndarray): _description_
        cut_left (float): _description_
        cut_right (float): _description_
    """
    # Constants init
    n_slices = len(profile)
    bin_width = (cut_right - cut_left) / n_slices
    inv_bin_width = 1. / bin_width
    const1 = (cut_left + bin_width * 0.5)
    const2 = (cut_right - bin_width * 0.5)

    profile[:] = 0.0

    for a in dt:
        if (a < const1) or (a > const2):
            continue

        fbin = (a - cut_left) * inv_bin_width
        ffbin = int(fbin)
        distToCenter = fbin - ffbin
        if distToCenter > 0.5:
            fffbin = int(fbin + 1)
        else:
            fffbin = int(fbin - 1)
        profile[ffbin] += (0.5 - distToCenter)
        profile[fffbin] += (0.5 + distToCenter)
# ---------------------------------------------------


# --------------- Similar to linear_interp_kick.cpp -----------------
@jit(nopython=True, nogil=True, fastmath=True, parallel=True, cache=True)
def linear_interp_kick(dt: np.ndarray, dE: np.ndarray, voltage: np.ndarray,
                       bin_centers: np.ndarray, charge: float,
                       acceleration_kick: float) -> None:
    """Interpolated kick method.

    Args:
        dt (np.ndarray): _description_
        dE (np.ndarray): _description_
        voltage (np.ndarray): _description_
        bin_centers (np.ndarray): _description_
        charge (float): _description_
        acceleration_kick (float): _description_
    """
    n_slices = len(bin_centers)
    inv_bin_width = (n_slices - 1) / (bin_centers[-1] - bin_centers[0])

    helper = np.empty(2 * (n_slices-1), dtype=np.float64)
    for i in prange(n_slices-1):
        helper[2*i] = charge * (voltage[i + 1] - voltage[i]) * inv_bin_width
        helper[2*i+1] = (charge * voltage[i] - bin_centers[i]
                         * helper[2*i]) + acceleration_kick

    for i in prange(len(dt)):
        fbin = int(np.floor((dt[i]-bin_centers[0])*inv_bin_width))
        if (fbin >= 0) and (fbin < n_slices - 1):
            dE[i] += dt[i] * helper[2*fbin] + helper[2*fbin+1]

# ---------------------------------------------------


# --------------- Similar to synchrotron_radiation.cpp -----------------
@jit(nopython=True, nogil=True, fastmath=True, parallel=True, cache=True)
def synchrotron_radiation(dE: np.ndarray, U0: float,
                          n_kicks: int, tau_z: float) -> None:
    """Apply SR

    Args:
        dE (np.ndarray): _description_
        U0 (float): _description_
        n_kicks (int): _description_
        tau_z (float): _description_
    """
    # Adjust inputs before the loop to reduce computations
    U0 = U0 / n_kicks
    tau_z = tau_z * n_kicks

    # SR damping constant, adjusted for better performance
    const_synch_rad = 1.0 - 2.0 / tau_z

    for i in prange(len(dE)):
        for _ in range(n_kicks):
            dE[i] = dE[i] * const_synch_rad - U0


@jit(nopython=True, nogil=True, fastmath=True, parallel=True, cache=True)
def synchrotron_radiation_full(dE: np.ndarray, U0: float,
                               n_kicks: int, tau_z: float,
                               sigma_dE: float, energy: float) -> None:
    """Apply SR with quantum excitation

    Args:
        dE (np.ndarray): _description_
        U0 (float): _description_
        n_kicks (int): _description_
        tau_z (float): _description_
        sigma_dE (float): _description_
        energy (float): _description_
    """

    # Adjust inputs before the loop to reduce computations
    U0 = U0 / n_kicks
    tau_z = tau_z * n_kicks

    const_quantum_exc = 2.0 * sigma_dE / np.sqrt(tau_z) * energy
    const_synch_rad = 1.0 - 2.0 / tau_z

    for i in prange(len(dE)):
        # rand_arr = np.random.normal(0.0, 1.0, size=n_kicks)
        for j in range(n_kicks):
            dE[i] = dE[i] * const_synch_rad + \
                const_quantum_exc * random.gauss(0.0, 1.0)-U0


# @jit(nopython=False, nogil=True, fastmath=False, parallel=False)
def set_random_seed(seed: int) -> None:
    """Set the seed of the RNG used in synchrotron radiation

    Args:
        seed (int): _description_
    """
    np.random.seed(seed)
    random.seed(seed)
# ---------------------------------------------------


# --------------- Similar to music_track.cpp -----------------
@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def music_track(dt: np.ndarray, dE: np.ndarray, induced_voltage: np.ndarray,
                array_parameters: np.ndarray, alpha: float, omega_bar: float,
                const: float, coeff1: float, coeff2: float,
                coeff3: float, coeff4: float) -> None:
    '''
    This function calculates the single-turn induced voltage and updates the
    energies of the particles.

    Parameters
    ----------
    beam_dt : float array
        Longitudinal coordinates [s]
    beam_dE : float array
        Initial energies [V]
    induced_voltage : float array
        array used to store the output of the computation
    array_parameters : float array
        See documentation in music.py
    n_macroparticles : int
        number of macro-particles
    alpha, omega_bar, cnst, coeff1, coeff2, coeff3, coeff4 : floats
        See documentation in music.py

    Returns
    -------
    induced_voltage : float array
        Computed induced voltage.
    beam_dE : float array
        Array of energies updated.
    '''

    indices_sorted = np.argsort(dt)
    dt = dt[indices_sorted]
    dE = dE[indices_sorted]

    # MuSiC algorithm
    dE[0] += induced_voltage[0]
    input_first_component = 1.0
    input_second_component = 0.0

    for i in range(len(dt) - 1):

        time_difference = dt[i + 1] - dt[i]

        exp_term = np.exp(-alpha * time_difference)
        cos_term = np.cos(omega_bar * time_difference)
        sin_term = np.sin(omega_bar * time_difference)

        product_first_component = exp_term * \
            ((cos_term + coeff1 * sin_term) * input_first_component
                + coeff2 * sin_term * input_second_component)
        product_second_component = exp_term * \
            (coeff3 * sin_term * input_first_component
                + (cos_term + coeff4 * sin_term) * input_second_component)

        induced_voltage[i + 1] = const * (0.5 + product_first_component)
        dE[i + 1] += induced_voltage[i + 1]

        input_first_component = product_first_component + 1.0
        input_second_component = product_second_component

    array_parameters[0] = input_first_component
    array_parameters[1] = input_second_component
    array_parameters[3] = dt[-1]


@jit(nopython=True, nogil=True, fastmath=True, cache=True)
def music_track_multiturn(dt: np.ndarray, dE: np.ndarray, induced_voltage: np.ndarray,
                          array_parameters: np.ndarray, alpha: float, omega_bar: float,
                          const: float, coeff1: float, coeff2: float,
                          coeff3: float, coeff4: float) -> None:
    """This function calculates the multi-turn induced voltage and updates the
    energies of the particles.
    Parameters and Returns as for music_track.

    Args:
        dt (np.ndarray): _description_
        dE (np.ndarray): _description_
        induced_voltage (np.ndarray): _description_
        array_parameters (np.ndarray): _description_
        alpha (float): _description_
        omega_bar (float): _description_
        const (float): _description_
        coeff1 (float): _description_
        coeff2 (float): _description_
        coeff3 (float): _description_
        coeff4 (float): _description_
    """
    indices_sorted = np.argsort(dt)
    dt = dt[indices_sorted]
    dE = dE[indices_sorted]
    time_difference_0 = dt[0] + array_parameters[2] - array_parameters[3]
    exp_term = np.exp(-alpha * time_difference_0)
    cos_term = np.cos(omega_bar * time_difference_0)
    sin_term = np.sin(omega_bar * time_difference_0)

    product_first_component = exp_term * (
        (cos_term + coeff1 * sin_term)
        * array_parameters[0]
        + coeff2 * sin_term * array_parameters[1])

    product_second_component = exp_term * (
        coeff3 * sin_term * array_parameters[0]
        + (cos_term + coeff4 * sin_term) * array_parameters[1])

    induced_voltage[0] = const * (0.5 + product_first_component)

    dE[0] += induced_voltage[0]
    input_first_component = product_first_component + 1.0
    input_second_component = product_second_component

    # MuSiC algorithm for the current turn

    for i in range(len(dt) - 1):

        time_difference = dt[i + 1] - dt[i]

        exp_term = np.exp(-alpha * time_difference)
        cos_term = np.cos(omega_bar * time_difference)
        sin_term = np.sin(omega_bar * time_difference)

        product_first_component = exp_term * \
            ((cos_term + coeff1 * sin_term) * input_first_component
                + coeff2 * sin_term * input_second_component)
        product_second_component = exp_term * \
            (coeff3 * sin_term * input_first_component
                + (cos_term + coeff4 * sin_term) * input_second_component)

        induced_voltage[i + 1] = const * (0.5 + product_first_component)
        dE[i + 1] += induced_voltage[i + 1]

        input_first_component = product_first_component + 1.0
        input_second_component = product_second_component

    array_parameters[0] = input_first_component
    array_parameters[1] = input_second_component
    array_parameters[3] = dt[-1]
# ---------------------------------------------------


# # --------------- Similar to fast_resonator.cpp -----------------
# @jit(nopython=True, nogil=True, fastmath=True, parallel=True, cache=True)
# def fast_resonator(R_S: np.ndarray, Q: np.ndarray, frequency_array: np.ndarray,
#                    frequency_R: np.ndarray, impedance: np.ndarray = None) -> np.ndarray:

#     if impedance is None:
#         impedance = np.zeros(len(frequency_array), dtype=np.complex128)

#     for freq in prange(1, len(frequency_array)):
#         impedance[freq] = 0.0
#         for i in range(len(R_S)):
#             impedance[freq] += R_S[i] / (1 + 1j * Q[i] * (frequency_array[freq] / frequency_R[i] -
#                                                           frequency_R[i] / frequency_array[freq]))

#     return impedance
# # ---------------------------------------------------

# --------------- Similar to fast_resonator.cpp -----------------
def fast_resonator(R_S: np.ndarray, Q: np.ndarray, frequency_array: np.ndarray,
                   frequency_R: np.ndarray, impedance: np.ndarray = None) -> np.ndarray:
    '''
    We're defining and calling a function internally due to issues
    dealing with parallelization and the allocation of the impedance array.
    '''


    if impedance is None:
        impedance = np.zeros(len(frequency_array), dtype=np.complex128)
    
    @jit(nopython=True, nogil=True, fastmath=True, parallel=True, cache=True)
    def calc_impedance(R_S: np.ndarray, Q: np.ndarray, frequency_array: np.ndarray,
                   frequency_R: np.ndarray, impedance: np.ndarray):
        for freq in prange(1, len(frequency_array)):
            for i in range(len(R_S)):
                impedance[freq] += R_S[i] / (1 + 1j * Q[i] * (frequency_array[freq] / frequency_R[i] -
                                                            frequency_R[i] / frequency_array[freq]))

    calc_impedance(R_S, Q, frequency_array, frequency_R, impedance)

    return impedance
# ---------------------------------------------------


# --------------- Similar to beam_phase.cpp -----------------
@jit(nopython=True, nogil=True, fastmath=True, parallel=True, cache=True)
def beam_phase(bin_centers: np.ndarray, profile: np.ndarray,
               alpha: float, omegarf: float,
               phirf: float, bin_size: float) -> float:
    scoeff = np.trapz(np.exp(alpha * (bin_centers))
                      * np.sin(omegarf * bin_centers + phirf)
                      * profile, dx=bin_size)
    ccoeff = np.trapz(np.exp(alpha * (bin_centers))
                      * np.cos(omegarf * bin_centers + phirf)
                      * profile, dx=bin_size)

    return scoeff / ccoeff


@jit(nopython=True, nogil=True, fastmath=True, parallel=True, cache=True)
def beam_phase_fast(bin_centers: np.ndarray, profile: np.ndarray,
                    omegarf: float, phirf: float, bin_size: float) -> float:
    scoeff = np.trapz(profile * np.sin(omegarf * bin_centers + phirf),
                      dx=bin_size)
    ccoeff = np.trapz(profile * np.cos(omegarf * bin_centers + phirf),
                      dx=bin_size)

    return scoeff / ccoeff
# ---------------------------------------------------


# --------------- Similar to sparse_histogram.cpp -----------------
@jit(nopython=True, nogil=True, fastmath=True, parallel=False)
def sparse_histogram(dt: np.ndarray, profile: np.ndarray,
                     cut_left: np.ndarray, cut_right: np.ndarray,
                     bunch_indexes: np.ndarray, n_slices_bucket: int) -> None:
    '''
    Optimised routine that calculates the histogram for a sparse beam
    Author: Juan F. Esteban Mueller, Danilo Quartullo, Alexandre Lasheen, Markus Schwarz
    '''
    # Only valid for cut_edges = edges
    inv_bucket_length = 1.0 / (cut_right[0] - cut_left[0])
    inv_bin_width = inv_bucket_length * n_slices_bucket

    profile[:] = 0.0

    for a in dt:
        if ((a < cut_left[0]) or (a > cut_right[-1])):
            continue
        # Find bucket in which the particle is and its index
        fbunch = int((a - cut_left[0]) * inv_bucket_length)
        i_bucket = int(bunch_indexes[fbunch])
        if i_bucket == -1:
            continue
        # Find the bin inside the corresponding bucket
        fbin = int((a - cut_left[i_bucket]) * inv_bin_width)
        profile[i_bucket, fbin] += 1.0
# ---------------------------------------------------


# --------------- Similar to tomoscope.cpp -----------------
@jit(nopython=True, nogil=True, fastmath=True, parallel=True, cache=True)
def distribution_from_tomoscope(dt: np.ndarray, dE: np.ndarray, probDistr: np.ndarray,
                                seed: int, profLen: int,
                                cutoff: float, x0: float, y0: float,
                                dtBin: float, dEBin: float) -> None:
    '''
    Generation of particle distribution from probability density
    Author: Helga Timko
    '''
    import bisect

    # Initialize random seed
    np.random.seed(seed)

    # Initialize variables
    cutoff2 = cutoff * cutoff
    dtMin = -1.0 * x0 * dtBin
    dEMin = -1.0 * y0 * dEBin

    # Calculate cumulative probability
    # profLen2 = np.uint32(profLen * profLen)
    cumulDistr = np.cumsum(probDistr)

    # Normalize probability distribution
    totProb = cumulDistr[-1]
    cumulDistr = cumulDistr / totProb

    # Generate particle coordinates

    n = 0
    while n < len(dt):
        randProb = np.random.random()
        m = bisect.bisect(cumulDistr, randProb)
        # for m in range(profLen2):
        #     if randProb < cumulDistr[m]:
        #         break
        i = int(m / profLen)
        k = m % profLen

        iPos = i + np.random.random() - 0.5
        kPos = k + np.random.random() - 0.5

        if ((iPos - x0) * (iPos - x0) + (kPos - y0) * (kPos - y0)) < cutoff2:
            dt[n] = dtMin + iPos * dtBin
            dE[n] = dEMin + kPos * dEBin
            n += 1
# ---------------------------------------------------




# Copyright 2016 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
**Phase-space variable conversions related to action. Single-RF case without
intensity effects is considered.**

:Authors: **Helga Timko**
'''

from __future__ import division

from builtins import range

import numpy as np
from scipy.special import ellipe, ellipk


def x(phimax):

    return np.sin(0.5 * phimax)


def x2(phimax):

    return np.sin(0.5 * phimax)**2


def action_from_phase_amplitude(x2):
    '''
    Returns the relative action for given oscillation amplitude in time.
    Action is normalised to the value at the separatrix, given in units of 1. 
    '''

    action = np.zeros(len(x2))

    indices = np.where(x2 != 1.)[0]
    indices0 = np.where(x2 == 1.)[0]
    action[indices] = (ellipe(x2[indices]) -
                       (1. - x2[indices]) * ellipk(x2[indices]))

    if indices0:

        action[indices0] = np.float(ellipe(x2[indices0]))

    return action


def tune_from_phase_amplitude(phimax):
    '''
    Find the tune w.r.t. the central synchrotron frequency corresponding to a
    given amplitude of synchrotron oscillations in phase 
    '''

    return 0.5 * np.pi / ellipk(x(phimax))


def phase_amplitude_from_tune(tune):
    '''
    Find the amplitude of synchrotron oscillations in phase corresponding to a
    given tune w.r.t. the central synchrotron frequency
    '''

    n = len(tune)
    phimax = np.zeros(n)

    for i in range(n):

        if tune[i] == 1.:

            phimax[i] = 0.

        elif tune[i] == 0.:

            phimax[i] = np.pi

        else:

            guess = 0.5 * np.pi
            difference = 0.25 * np.pi
            k = 0

            while np.fabs(tune[i] - tune_from_phase_amplitude(guess)) / tune[i] \
                    > 0.001 and np.fabs(difference / guess) > 1.e-10:

                guess += np.sign(tune_from_phase_amplitude(guess)
                                 - tune[i]) * difference
                difference *= 0.5
                k += 1
                if k > 100:
                    # PhaseSpaceError
                    raise RuntimeError("Exceeded maximum number of iterations in phase_amplitude_from_tune()!")

            phimax[i] = guess

    return phimax


def oscillation_amplitude_from_coordinates(Ring, RFStation, dt, dE,
                                           timestep=0, Np_histogram=None):
    '''
    Returns the oscillation amplitude in time for given particle coordinates,
    assuming single-harmonic RF system and no intensity effects. 
    Optional: RF parameters at a given timestep (default = 0) are used.
    Optional: Number of points for histogram output
    '''

    omega_rf = RFStation.omega_rf[0, timestep]
    phi_rf = RFStation.phi_rf[0, timestep]
    phi_s = RFStation.phi_s[timestep]
    eta = RFStation.eta_0[0]
    T0 = Ring.t_rev[0]
    V = RFStation.voltage[0, 0]
    beta_sq = RFStation.beta[0]**2
    E = RFStation.energy[0]
    const = eta * T0 * omega_rf / (2. * V * beta_sq * E)

    dtmax = np.fabs(np.arccos(np.cos(omega_rf * dt + phi_rf) + const * dE**2)
                    - phi_rf - phi_s) / omega_rf

    if Np_histogram is not None:

        histogram, bins = np.histogram(dtmax, Np_histogram, (0,
                                                             np.pi / omega_rf))
        histogram = np.double(histogram) / np.sum(histogram[:])
        bin_centres = 0.5 * (bins[0:-1] + bins[1:])

        return dtmax, bin_centres, histogram

    else:

        return dtmax


def action_from_oscillation_amplitude(RFStation, dtmax, timestep=0,
                                      Np_histogram=None):
    '''
    Returns the relative action for given oscillation amplitude in time,
    assuming single-harmonic RF system and no intensity effects.
    Action is normalised to the value at the separatrix, given in units of 1. 
    Optional: RF parameters at a given timestep (default = 0) are used.
    Optional: Number of points for histogram output
    '''

    omega_rf = RFStation.omega_RF[0, timestep]
    xx = x2(omega_rf * dtmax)
    action = np.zeros(len(xx))

    indices = np.where(xx != 1.)[0]
    indices0 = np.where(xx == 1.)[0]
    action[indices] = (ellipe(xx[indices]) -
                       (1. - xx[indices]) * ellipk(xx[indices]))
    if indices0:

        action[indices0] = np.float(ellipe(xx[indices0]))

    if Np_histogram is not None:

        histogram, bins = np.histogram(action, Np_histogram, (0, 1))
        histogram = np.double(histogram) / np.sum(histogram[:])
        bin_centres = 0.5 * (bins[0:-1] + bins[1:])

        return action, bin_centres, histogram

    else:

        return action




# Copyright 2016 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3), 
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities 
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
**Module to plot different bunch features**

:Authors: **Helga Timko**, **Danilo Quartullo**
'''

from __future__ import division
import numpy as np
import matplotlib.pyplot as plt
from ..trackers.utilities import separatrix
from matplotlib.layout_engine import PlaceHolderLayoutEngine



def plot_long_phase_space(Ring, RFStation, Beam, xmin,
                          xmax, ymin, ymax, xunit = 's', sampling = 1, 
                          separatrix_plot = False, histograms_plot = True, 
                          dirname = 'fig', show_plot = False, alpha = 1, color = 'b'):
    """
    Plot of longitudinal phase space. Optional use of histograms and separatrix.
    Choice of units: xunit = s, rad.
    For large amount of data, use "sampling" to plot a fraction of the data.
    """

    # Conversion from particle arrival time to RF phase
    if xunit == 'rad':
        omega_RF = RFStation.omega_rf[0,RFStation.counter[0]]
        phi_RF = RFStation.phi_rf[0,RFStation.counter[0]]

    # Definitions for placing the axes
    left, width = 0.115, 0.63
    bottom, height = 0.115, 0.63
    bottom_h = left_h = left+width+0.03
    
    rect_scatter = [left, bottom, width, height]
    rect_histx = [left, bottom_h, width, 0.2]
    rect_histy = [left_h, bottom, 0.2, height]

    # Prepare plot
    fig = plt.figure(1)
    fig.set_layout_engine(PlaceHolderLayoutEngine(False,False))
    fig.set_size_inches(8,8)
    axScatter = plt.axes(rect_scatter)
    axHistx = plt.axes(rect_histx)
    axHisty = plt.axes(rect_histy)
    
    # Main plot: longitudinal distribution
    indlost = np.where( Beam.id[::sampling] == 0 )[0] # particles lost
    indalive = np.where( Beam.id[::sampling] != 0 )[0] # particles transmitted
    if xunit == 's':
        axScatter.set_xlabel(r"$\Delta t$ [s]")
        axScatter.scatter(Beam.dt[indalive], Beam.dE[indalive], 
                          s=1, edgecolor='none', alpha=alpha, color=color)
        if len(indlost) > 0:
            axScatter.scatter(Beam.dt[indlost], Beam.dE[indlost], color='0.5',
                              s=1, edgecolor='none')
    elif xunit == 'rad':
        axScatter.set_xlabel(r"$\varphi$ [rad]")
        axScatter.scatter(omega_RF*Beam.dt[indalive] + phi_RF, 
                          Beam.dE[indalive], s=1, edgecolor='none', 
                          alpha=alpha, color=color)
        if len(indlost) > 0:
            axScatter.scatter(omega_RF*Beam.dt[indlost] + phi_RF, 
                              Beam.dE[indlost], color='0.5', s=1, 
                              edgecolor='none')
    axScatter.set_ylabel(r"$\Delta$E [eV]")
    axScatter.yaxis.labelpad = 1     
        
    axScatter.set_xlim(xmin, xmax)
    axScatter.set_ylim(ymin, ymax)
    
    axScatter.ticklabel_format(style='sci', axis='y', scilimits=(0,0))
    plt.figtext(0.95,0.95,'%d turns' %RFStation.counter[0], 
                fontsize=16, ha='right', va='center') 
            
    # Separatrix
    if separatrix_plot:
        x_sep = np.linspace(float(xmin), float(xmax), 1000)
        if xunit == 's':
            y_sep = separatrix(Ring, RFStation, x_sep)
        elif xunit == 'rad':
            y_sep = separatrix(Ring, RFStation, (x_sep - phi_RF)/omega_RF)
        axScatter.plot(x_sep, y_sep, 'r')
        axScatter.plot(x_sep, - y_sep, 'r')       
   
    # Phase and momentum histograms
    if histograms_plot:
        xbin = (xmax - xmin)/200.
        xh = np.arange(xmin, xmax + xbin, xbin)
        ybin = (ymax - ymin)/200.
        yh = np.arange(ymin, ymax + ybin, ybin)
      
        if xunit == 's':
            axHistx.hist(Beam.dt[::sampling], bins=xh, histtype='step', color=color)
        elif xunit == 'rad':
            axHistx.hist(omega_RF*Beam.dt[::sampling] + phi_RF, bins=xh, histtype='step', color=color)       
        axHisty.hist(Beam.dE[::sampling], bins=yh, histtype='step', 
                     orientation='horizontal', color=color)
        axHistx.ticklabel_format(style='sci', axis='y', scilimits=(0,0))
        axHisty.ticklabel_format(style='sci', axis='x', scilimits=(0,0))
        axHistx.axes.get_xaxis().set_visible(False)
        axHisty.axes.get_yaxis().set_visible(False)
        axHistx.set_xlim(xmin, xmax)
        axHisty.set_ylim(ymin, ymax)
        labels = axHisty.get_xticklabels()
        for label in labels:
            label.set_rotation(-90)
                     
    # Output plot
    if show_plot:
        plt.show()
    else:
        fign = dirname +'/long_distr_'"%d"%RFStation.counter[0]+'.png'
        plt.savefig(fign)
    plt.clf()



def plot_bunch_length_evol(RFStation, h5data, output_freq = 1, 
                           dirname = 'fig', show_plot = False):
    """
    Plot of r.m.s. 4-sigma bunch length [s] as a function of time.
    """

    # Time step of plotting
    time_step = RFStation.counter[0]
    

    # Get bunch length data in metres or seconds
    if output_freq < 1:
        output_freq = 1
    ndata = int(time_step/output_freq)
    t = output_freq*np.arange(ndata)  
    bl = np.array(h5data["/Beam/sigma_dt"][0:ndata], dtype = np.double)
    bl *= 4
    bl[time_step:] = np.nan
    
    # Plot
    fig = plt.figure(1)
    fig.set_size_inches(8,6)
    ax = plt.axes()
    ax.plot(t, bl, '.')
    ax.set_xlabel(r"No. turns [T$_0$]")
    ax.set_ylabel (r"Bunch length, $\Delta t_{4\sigma}$ r.m.s. [s]")
    if time_step > 100000:
        ax.ticklabel_format(style='sci', axis='x', scilimits=(0,0))

    # Output plot
    if show_plot:
        plt.show()
    else:
        fign = dirname +'/bunch_length.png'
        plt.savefig(fign)
    plt.clf()

    

def plot_bunch_length_evol_gaussian(RFStation, Profile, h5data, 
                                    output_freq = 1, dirname = 'fig', show_plot = False):

    """
    Plot of Gaussian 4-sigma bunch length [s] as a function of time.
    Requires profile.
    """

    # Time step of plotting
    time_step = RFStation.counter[0]

    # Get bunch length data in metres or nanoseconds
    if output_freq < 1:
        output_freq = 1
    ndata = int(time_step/output_freq)
    t = output_freq*np.arange(ndata)
    
    bl = np.array(h5data["/Beam/bunch_length"][0:ndata], dtype=np.double)
    
    bl[time_step:] = np.nan

    # Plot
    fig = plt.figure(1)
    fig.set_size_inches(8,6)
    ax = plt.axes()
    ax.plot(t, bl, '.')
    ax.set_xlabel(r"No. turns [T$_0$]")
    ax.set_ylabel (r"Bunch length, $\Delta t_{4\sigma}$ Gaussian fit [s]")
    if time_step > 100000:
        ax.ticklabel_format(style='sci', axis='x', scilimits=(0,0))
   
    # Output plot
    if show_plot:
        plt.show()
    else:
        fign = dirname +'/bunch_length_Gaussian.png'
        plt.savefig(fign)
    plt.clf()

    

def plot_position_evol(RFStation, h5data, output_freq = 1, 
                       style = '.', dirname = 'fig', show_plot = False):
 
    # Time step of plotting
    time_step = RFStation.counter[0]
    
    # Get position data [s] 
    if output_freq < 1:
        output_freq = 1
    ndata = int(time_step/output_freq)
    t = output_freq*np.arange(ndata)    
    pos = np.array(h5data["/Beam/mean_dt"][0:ndata], dtype = np.double)          
    pos[time_step:] = np.nan 
 
    # Plot 
    fig = plt.figure(1)
    fig.set_size_inches(8,6)
    ax = plt.axes()
    ax.plot(t, pos, style) 
    ax.set_xlabel(r"No. turns [T$_0$]") 
    ax.set_ylabel (r"Bunch mean position, $<\Delta t>$ [s]") 
    if time_step > 100000:
        ax.ticklabel_format(style='sci', axis='x', scilimits=(0,0))
     
    # Output plot
    if show_plot:
        plt.show()
    else:
        fign = dirname+'/bunch_mean_position.png'
        plt.savefig(fign)
    plt.clf() 



def plot_energy_evol(RFStation, h5data, output_freq = 1, style = '.', 
                     dirname = 'fig', show_plot = False):
 
    # Time step of plotting
    time_step = RFStation.counter[0]

    # Get position data in metres or nanoseconds 
    if output_freq < 1:
        output_freq = 1
    ndata = int(time_step/output_freq)
    t = output_freq*np.arange(ndata)  
    nrg = np.array(h5data["/Beam/mean_dE"][0:ndata], dtype = np.double)          
    nrg[time_step:] = np.nan 
 
    # Plot 
    fig = plt.figure(1)
    fig.set_size_inches(8,6)
    ax = plt.axes()
    ax.plot(t, nrg, style) 
    ax.set_xlabel(r"No. turns [T$_0$]") 
    ax.set_ylabel (r"Bunch mean energy, $<\Delta E>$ [eV]") 
    ax.ticklabel_format(style='sci', axis='y', scilimits=(0,0))
    if time_step > 100000:
        ax.ticklabel_format(style='sci', axis='x', scilimits=(0,0))
     
    # Output plot
    if show_plot:
        plt.show()
    else:
        fign = dirname+'/bunch_mean_energy.png'
        plt.savefig(fign)
    plt.clf() 


def plot_transmitted_particles(RFStation, h5data, output_freq = 1, 
                               style = '.', dirname = 'fig', show_plot = False):
 
    # Time step of plotting
    time_step = RFStation.counter[0]

    # Get position data in metres or nanoseconds 
    if output_freq < 1:
        output_freq = 1
    ndata = int(time_step/output_freq)
    t = output_freq*np.arange(ndata)
    prtcls = np.array(h5data["/Beam/n_macroparticles_alive"][0:ndata], 
                      dtype = np.double)          
    prtcls[time_step:] = np.nan 
 
    # Plot 
    plt.figure(1, figsize=(8,6)) 
    ax = plt.axes()
    ax.plot(t, prtcls, style) 
    ax.set_xlabel(r"No. turns [T$_0$]") 
    ax.set_ylabel (r"Transmitted macro-particles [1]") 
    ax.ticklabel_format(style='sci', axis='y', scilimits=(0,0))
    if time_step > 100000:
        ax.ticklabel_format(style='sci', axis='x', scilimits=(0,0))

    # Output plot
    if show_plot:
        plt.show()
    else:
        fign = dirname+'/bunch_transmitted_particles.png'
        plt.savefig(fign)
    plt.clf() 



# Copyright 2016 CERN. This software is distributed under the
# terms of the GNU General Public Licence version 3 (GPL Version 3),
# copied verbatim in the file LICENCE.md.
# In applying this licence, CERN does not waive the privileges and immunities
# granted to it by virtue of its status as an Intergovernmental Organization or
# submit itself to any jurisdiction.
# Project website: http://blond.web.cern.ch/

'''
**Module to wrap tracking into an iterator with option to call
user specified functions every n turns**

:Authors: **Simon Albright**
'''
# Futurisation
from __future__ import annotations

# General imports
from typing import TYPE_CHECKING


if TYPE_CHECKING:
    from typing import Iterable, List, Callable, Protocol, Any, Self, Tuple

    class Trackable(Protocol):
        def track(self) -> None:
            ...

    class Predicate(Protocol):
        def __call__(self, _map: Iterable[Trackable], turn_number: int,
                     *args: Any, **kwargs: Any):
            ...



class TrackIteration:
    '''
    Class to provide an iterator for tracking with an option to run passed
    functions every n turns

    Parameters
    ----------

    track_map : iterable of objects
        Each object will be called on every turn with object.track()
    init_turn : integer
        The turn number tracking will start from, only used to initialise
        a turn counter
    final_turn : integer
        The last turn number to track next(TrackIteration) will raise
        StopIteration when turn_number == final_turn

    Attributes
    ----------
    function_list : List of functions to be called with specified interval
    '''


    def __init__(self, track_map: Iterable[Trackable], init_turn: int = 0,
                 final_turn: int = -1):

        if not all((hasattr(m, 'track') for m in track_map)):
            raise AttributeError("All map objects must be trackable")

        self._map = list(track_map)
        if isinstance(init_turn, int):
            self.turn_number = init_turn
        else:
            raise TypeError("init_turn must be an integer")

        if isinstance(final_turn, int):
            self._final_turn = final_turn
        else:
            raise TypeError("final_turn must be an integer")

        self.function_list: List[Tuple[Predicate, int]] = []


    def _track_turns(self, n_turns):
        '''
        Function to track for specified number of turns
        calls next() function n_turns times
        '''

        for i in range(n_turns):
            next(self)


    def add_function(self, predicate: Predicate, repetion_rate: int,
                     *args: Any, **kwargs: Any):
        '''
        Takes a user defined callable and calls it every repetion_rate
        number of turns with predicate(track_map, turn_number, *args, **kwargs)
        '''

        self.function_list.append((self._partial(predicate, *args, **kwargs),
                                  repetion_rate))


    def __next__(self) -> int:
        '''
        First raises StopIteration if turn_number == final_turn

        Next calls track() from each element in trackMap list and raises
        StopIteration if no more turns available

        Finally iterates over each function specified in add_function
        and calls them with predicate(trackMap, turn_number) if
        turn_number % repetitionRate == 0
        '''

        if self.turn_number == self._final_turn:
            raise StopIteration

        try:
            for m in self._map:
                m.track()
        except IndexError:
            raise StopIteration

        self.turn_number += 1

        for func, rate in self.function_list:
            if self.turn_number % rate == 0:
                func(self._map, self.turn_number)

        return self.turn_number


    def __iter__(self) -> Self:
        '''
        returns self
        '''

        return self


    def __call__(self, n_turns: int = 1) -> int:
        '''
        Makes object callable with option to specify number of tracked turns
        default tracks 1 turn
        '''

        self._track_turns(n_turns)
        return self.turn_number


    def _partial(self, predicate: Callable, *args, **kwargs) -> Callable:
        '''
        reimplementation of functools.partial to prepend
        rather than append to *args
        '''

        def part_func(_map, turn):
            return predicate(_map, turn, *args, **kwargs)

        return part_func
